from collections.abc import Sequence
from dataclasses import dataclass
from math import prod

from xdsl.builder import ImplicitBuilder
from xdsl.context import Context
from xdsl.dialects import arith, builtin, memref, stencil, tensor, varith
from xdsl.dialects.builtin import (
    AnyTensorType,
    DenseIntOrFPElementsAttr,
    FloatAttr,
    IndexType,
    IntegerAttr,
    IntegerType,
    MemRefType,
    ModuleOp,
    TensorType,
)
from xdsl.dialects.csl import csl_stencil
from xdsl.dialects.experimental import dmp
from xdsl.ir import (
    Attribute,
    Block,
    BlockArgument,
    Operation,
    OpResult,
    Region,
    SSAValue,
)
from xdsl.irdl import Operand
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    GreedyRewritePatternApplier,
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    op_type_rewrite_pattern,
)
from xdsl.rewriter import InsertPoint
from xdsl.transforms.experimental.stencil_tensorize_z_dimension import (
    BackpropagateStencilShapes,
)
from xdsl.transforms.varith_transformations import (
    ConvertArithToVarithPass,
    ConvertVarithToArithPass,
)
from xdsl.utils.hints import isa


def get_stencil_access_operands(op: Operand) -> set[Operand]:
    """
    Returns the symbols of all stencil accessess by op and all its dependencies.
    """
    res: set[Operand] = set()
    frontier: set[Operand] = {op}
    next: set[Operand] = set()
    while len(frontier) > 0:
        for o in frontier:
            if isinstance(o, OpResult):
                if isinstance(o.op, csl_stencil.AccessOp):
                    res.add(o.op.op)
                else:
                    next.update(o.op.operands)
        frontier = next
        next = set()

    return res


def _get_prefetch_buf_idx(op: stencil.ApplyOp) -> int | None:
    # calculate memory cost of all prefetch operands
    def get_prefetch_overhead(o: OpResult):
        assert isa(o.type, TensorType[Attribute])
        return prod(o.type.get_shape())

    candidate_prefetches = [
        (get_prefetch_overhead(o), o)
        for o in op.operands
        if isinstance(o, OpResult) and isinstance(o.op, csl_stencil.PrefetchOp)
    ]
    if len(candidate_prefetches) == 0:
        return

    # select the prefetch with the biggest communication overhead to be fused with matched stencil.apply
    prefetch = max(candidate_prefetches, key=lambda x: x[0])[1]
    return op.operands.index(prefetch)


def _get_apply_op(op: Operation) -> stencil.ApplyOp | None:
    """
    Return the enclosing csl_wrapper.module
    """
    parent_op = op.parent_op()
    while parent_op:
        if isinstance(parent_op, stencil.ApplyOp):
            return parent_op
        parent_op = parent_op.parent_op()
    return None


@dataclass(frozen=True)
class ConvertAccessOpPattern(RewritePattern):
    """
    Rebuilds stencil.access by csl_stencil.access which operates on prefetched accesses.

    stencil.access operates on stencil.temp types found at arg_index
    csl_stencil.access operates on memref< num_neighbors x tensor< buf_size x data_type >> found at last arg index

    Note: This is intended to be called in a nested pattern rewriter, such that the above precondition is met.
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: stencil.AccessOp, rewriter: PatternRewriter, /):
        assert len(op.offset) == 2
        if isa(op.temp.type, AnyTensorType):
            res_type = TensorType(
                op.temp.type.get_element_type(), op.temp.type.get_shape()[1:]
            )
        else:
            assert isa(op.res.type, AnyTensorType)
            res_type = op.res.type
        rewriter.replace_matched_op(
            new_access_op := csl_stencil.AccessOp(
                op=op.temp,
                offset=op.offset,
                offset_mapping=op.offset_mapping,
                result_type=res_type,
            )
        )

        # The stencil-tensorize-z-dimension pass inserts tensor.ExtractSliceOps after stencil.access to remove ghost cells.
        # Since ghost cells are not prefetched, these ops can be removed again. Check if the ExtractSliceOp
        # has no other effect and if so, remove both.
        if (
            isinstance(
                use := new_access_op.result.get_user_of_unique_use(),
                tensor.ExtractSliceOp,
            )
            and use.static_sizes.get_values() == res_type.get_shape()
            and len(use.offsets) == 0
            and len(use.sizes) == 0
            and len(use.strides) == 0
        ):
            rewriter.replace_op(use, [], new_results=[new_access_op.result])


@dataclass
class ConvertSwapToPrefetchPattern(RewritePattern):
    """
    Translates dmp.swap to csl_stencil.prefetch
    """

    num_chunks: int = 1

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: dmp.SwapOp, rewriter: PatternRewriter, /):
        # remove op if it contains no swaps
        if len(op.swaps) == 0:
            rewriter.erase_matched_op(False)
            return

        assert all(len(swap.size) == 3 for swap in op.swaps), (
            "currently only 3-dimensional stencils are supported"
        )

        assert all(swap.size[:2] == (1, 1) for swap in op.swaps), (
            "invoke dmp to decompose from (x,y,z) to (1,1,z)"
        )

        # check that size is uniform
        uniform_size = op.swaps.data[0].size[2]
        assert all(swap.size[2] == uniform_size for swap in op.swaps), (
            "all swaps need to be of uniform size"
        )

        assert (MemRefType.constr() | stencil.StencilTypeConstr).verifies(
            op.input_stencil.type
        )
        assert isa(
            t_type := op.input_stencil.type.get_element_type(), TensorType[Attribute]
        )
        assert op.strategy.comm_layout() is not None, (
            f"topology on {type(op)} is not given"
        )

        # when translating swaps, remove third dimension
        prefetch_op = csl_stencil.PrefetchOp(
            input_stencil=op.input_stencil,
            topo=op.strategy.comm_layout(),
            num_chunks=IntegerAttr(self.num_chunks, 64),
            swaps=[
                csl_stencil.ExchangeDeclarationAttr(swap.neighbor[:2])
                for swap in op.swaps
            ],
            result_type=TensorType(
                t_type.get_element_type(),
                (len(op.swaps), uniform_size),
            ),
        )

        # if the rewriter needs a result, use `input_stencil` as a drop-in replacement
        # prefetch_op produces a result that needs to be handled separately
        # note, that only un-bufferized dmp.swaps produce a result
        rewriter.replace_matched_op(
            prefetch_op, new_results=[op.input_stencil] if op.swapped_values else []
        )

        # uses have to be retrieved *before* the loop because of the rewriting happening inside the loop
        uses = list(op.input_stencil.uses)

        # csl_stencil.prefetch, unlike dmp.swap, has a return value. This is added as the last arg
        # to stencil.apply, before rebuilding the op and replacing stencil.access ops by csl_stencil.access ops
        # that reference the prefetched buffers (note, this is only done for neighbor accesses)
        for use in uses:
            if not isinstance(use.operation, stencil.ApplyOp):
                continue
            apply_op = use.operation

            # arg_idx points to the stencil.temp type whose data is prefetched in a separate buffer
            arg_idx = apply_op.args.index(op.input_stencil)
            field_block_arg = apply_op.region.block.args[arg_idx]

            # add the prefetched buffer as the last arg to stencil.access
            prefetch_block_arg = apply_op.region.block.insert_arg(
                prefetch_op.result.type, len(apply_op.args)
            )
            field_block_arg.replace_by_if(
                prefetch_block_arg,
                lambda use: isinstance(use.operation, stencil.AccessOp)
                and tuple(use.operation.offset) != (0, 0),
            )

            # rebuild stencil.apply op
            r_types = apply_op.result_types
            assert isa(r_types, Sequence[stencil.TempType[Attribute]])
            new_apply_op = stencil.ApplyOp.build(
                operands=[[*apply_op.args, prefetch_op.result], apply_op.dest],
                regions=[apply_op.detach_region(apply_op.region)],
                result_types=[r_types],
                properties=apply_op.properties,
                attributes=apply_op.attributes,
            )
            rewriter.replace_op(apply_op, new_apply_op)


def split_ops(
    ops: Sequence[Operation], buf: BlockArgument
) -> tuple[Sequence[Operation], Sequence[Operation]]:
    """
    Returns a split of `ops` into an `(a,b)` tuple, such that:

    - `a` contains neighbour accesses to `buf` plus the minumum set of instructions to reduce the accessed data to 1 thing
    - `b` contains everything else

    If no valid split can be found, return `(ops, [])`.

    This function does not attempt to arithmetically re-structure the computation to obtain a good split. To do this,
    `RestructureSymmetricReductionPattern()` may be executed first.
    """
    a: list[Operation] = []
    b: list[Operation] = []
    rem: list[Operation] = []
    for op in ops:
        if isinstance(op, csl_stencil.AccessOp):
            (b, a)[op.op == buf].append(op)
        elif isinstance(op, arith.ConstantOp):
            a.append(op)
        else:
            rem.append(op)

    # loop until we can make no more changes, or until only 1 thing computed in `a` is used outside of it
    has_changes = True
    while (
        len(
            # ops in `a` whose results are used outside of `a`
            a_exports := set(
                op
                for op in a
                for result in op.results
                for use in result.uses
                if use.operation not in a
            )
        )
        > 1
        and has_changes
    ):
        has_changes = False

        # find ops that directly depend on `a` but are not themselves in `a`
        for exp in a_exports:
            for result in exp.results:
                for use in result.uses:
                    # op is only movable if *all* operands are already in `a` (and it hasn't been moved yet)
                    if (op := use.operation) in rem and all(
                        x.op in a for x in op.operands if isinstance(x, OpResult)
                    ):
                        has_changes = True
                        a.append(use.operation)
                        rem.remove(use.operation)

    # find constants in `a` needed outside of `a`
    cnst_exports = tuple(
        cnst for cnst in a_exports if isinstance(cnst, arith.ConstantOp)
    )

    # `a` exports one value plus any number of constants - duplicate exported constants and return op split
    if len(a_exports) == 1 + len(cnst_exports):
        recv_chunk_ops, done_exch_ops = list[Operation](), list[Operation]()
        for op in ops:
            if op in a:
                recv_chunk_ops.append(op)
                if op in cnst_exports:
                    assert isinstance(op, arith.ConstantOp)
                    # create a copy of the constant in the second region
                    done_exch_ops.append(cln := op.clone())
                    # rewire ops of the second region to use the copied constant
                    op.result.replace_by_if(
                        cln.result,
                        lambda use: use.operation in b or use.operation in rem,
                    )
            else:
                done_exch_ops.append(op)

        return recv_chunk_ops, done_exch_ops

    # fallback
    # always place `stencil.return` in second block
    return ops[:-1], [ops[-1]]


@dataclass(frozen=True)
class SplitVarithOpPattern(RewritePattern):
    """
    Splits a varith op into two, depending on whether the operands holds stencil accesses to `buf` (only)
    or any other accesses.

    This pass is intended to be run with `buf` set to the block arg indicating data received from neighbours.
    """

    buf: BlockArgument

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: varith.VarithOp, rewriter: PatternRewriter, /):
        if not (apply := _get_apply_op(op)) or not (
            buf_idx := _get_prefetch_buf_idx(apply)
        ):
            return
        buf = apply.region.block.args[buf_idx]
        buf_accesses, others = list[SSAValue](), list[SSAValue]()

        for arg in op.args:
            accs = get_stencil_access_operands(arg)
            (others, buf_accesses)[buf in accs and len(accs) == 1].append(arg)

        if len(others) > 0 and len(buf_accesses) > 0:
            rewriter.replace_matched_op(
                [
                    n_op := type(op)(*buf_accesses),
                    type(op)(n_op, *others),
                ]
            )


@dataclass(frozen=True)
class ConvertApplyOpPattern(RewritePattern):
    """
    Fuses a `csl_stencil.prefetch` and a `stencil.apply` to build a `csl_stencil.apply`.

    If there are several candidate prefetch ops, the one with the largest result buffer
    size is selected.
    The selection is greedy, and could in the future be expanded into a more global
    selection optimising for minimal prefetch overhead across multiple apply ops.
    """

    num_chunks: int = 1
    """
    Number of chunks into which communication and computation should be split.
    Effectively, the number of times `csl_stencil.apply.receive_chunk` will be executed
    and the tensor sizes it handles.
    Higher values may increase compute overhead but reduce size of communication buffers
    when lowered.
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: stencil.ApplyOp, rewriter: PatternRewriter, /):
        if not (prefetch_idx := _get_prefetch_buf_idx(op)):
            return

        # select the prefetch with the biggest communication overhead to be fused with matched stencil.apply
        prefetch = op.operands[prefetch_idx]
        assert isinstance(prefetch, OpResult)
        assert isinstance(prefetch.op, csl_stencil.PrefetchOp)
        field_idx = op.operands.index(prefetch.op.input_stencil)
        assert isinstance(prefetch.op, csl_stencil.PrefetchOp)
        assert isa(prefetch.type, TensorType[Attribute])
        field_op_arg = prefetch.op.input_stencil

        # add empty tensor before op to be used as `accumulator`
        # this could potentially be re-used if we have one of the same size lying around
        accumulator = tensor.EmptyOp(
            (),
            TensorType(prefetch.type.get_element_type(), prefetch.type.get_shape()[1:]),
        )
        rewriter.insert_op(accumulator, InsertPoint.before(op))

        # run pass (on this apply's region only) to consume data from `prefetch` accesses first
        # find varith ops and split according to neighbour data
        PatternRewriteWalker(
            SplitVarithOpPattern(op.region.block.args[prefetch_idx]),
            apply_recursively=False,
            listener=rewriter,
        ).rewrite_region(op.region)

        # determine how ops should be split across the two regions
        chunk_region_ops, done_exchange_ops = split_ops(
            list(op.region.block.ops), op.region.block.args[prefetch_idx]
        )

        # fetch what receive_chunk is computing for
        if isinstance(chunk_region_ops[-1], stencil.ReturnOp):
            chunk_res = chunk_region_ops[-1].operands[0]
        else:
            chunk_res = chunk_region_ops[-1].results[0]

        # after region split, check which block args (from the old ops block) are being accessed in each of the new regions
        # ignore accesses block args which already are part of the region's required signature
        chunk_region_used_block_args = sorted(
            set(
                x
                for o in chunk_region_ops
                for x in o.operands
                if isinstance(x, BlockArgument) and x.index != prefetch_idx
            ),
            key=lambda b: b.index,
        )
        done_exchange_used_block_args = sorted(
            set(
                x
                for o in done_exchange_ops
                for x in o.operands
                if isinstance(x, BlockArgument) and x.index != field_idx
            ),
            key=lambda b: b.index,
        )

        # set up region signatures, comprising fixed and optional args - see docs on `csl_stencil.apply` for details
        chunk_region_args = [
            # required arg 0: slice of type(%prefetch)
            TensorType(
                prefetch.type.get_element_type(),
                (
                    len(prefetch.op.swaps),
                    prefetch.type.get_shape()[1] // self.num_chunks,
                ),
            ),
            # required arg 1: %offset
            IndexType(),
            # required arg 2: %accumulator
            accumulator.tensor.type,
            # optional args: as needed by the ops
            *[a.type for a in chunk_region_used_block_args],
        ]
        done_exchange_args = [
            # required arg 0: stencil.temp to access own data
            field_op_arg.type,
            # required arg 1: %accumulator
            accumulator.tensor.type,
            # optional args: as needed by the ops
            *[a.type for a in done_exchange_used_block_args],
        ]

        # set up two regions
        receive_chunk = Region(Block(arg_types=chunk_region_args))
        done_exchange = Region(Block(arg_types=done_exchange_args))

        # translate old to new block arg index for optional args
        chunk_region_oprnd_table = dict[Operand, Operand](
            (old, receive_chunk.block.args[idx])
            for idx, old in enumerate(chunk_region_used_block_args, start=3)
        )
        done_exchange_oprnd_table = dict[Operand, Operand](
            (old, done_exchange.block.args[idx])
            for idx, old in enumerate(done_exchange_used_block_args, start=2)
        )

        # add translation from old to new arg index for non-optional args - note, access
        # to accumulator must be handled separately below
        chunk_region_oprnd_table[op.region.block.args[prefetch_idx]] = (
            receive_chunk.block.args[0]
        )
        done_exchange_oprnd_table[op.region.block.args[field_idx]] = (
            done_exchange.block.args[0]
        )
        done_exchange_oprnd_table[chunk_res] = done_exchange.block.args[1]

        # detach ops from old region
        for o in op.region.block.ops:
            op.region.block.detach_op(o)

        # add operations from list to receive_chunk, use translation table to rebuild operands
        for o in chunk_region_ops:
            if isinstance(o, stencil.ReturnOp | csl_stencil.YieldOp):
                break
            o.operands = [chunk_region_oprnd_table.get(x, x) for x in o.operands]
            rewriter.insert_op(o, InsertPoint.at_end(receive_chunk.block))

        # put `chunk_res` into `accumulator` (using tensor.insert_slice) and yield the result
        rewriter.insert_op(
            [
                insert_slice_op := tensor.InsertSliceOp.get(
                    source=chunk_res,
                    dest=receive_chunk.block.args[2],
                    offsets=(receive_chunk.block.args[1],),
                    static_sizes=(prefetch.type.get_shape()[1] // self.num_chunks,),
                ),
                csl_stencil.YieldOp(insert_slice_op.result),
            ],
            InsertPoint.at_end(receive_chunk.block),
        )

        # add operations from list to done_exchange, use translation table to rebuild operands
        for o in done_exchange_ops:
            o.operands = [done_exchange_oprnd_table.get(x, x) for x in o.operands]
            rewriter.insert_op(o, InsertPoint.at_end(done_exchange.block))
            if isinstance(o, stencil.ReturnOp):
                rewriter.replace_op(o, csl_stencil.YieldOp(*o.operands))

        rewriter.replace_matched_op(
            csl_stencil.ApplyOp(
                operands=[
                    field_op_arg,
                    accumulator,
                    [op.operands[a.index] for a in chunk_region_used_block_args],
                    [op.operands[a.index] for a in done_exchange_used_block_args],
                    op.dest,
                ],
                properties={
                    "swaps": prefetch.op.swaps,
                    "topo": prefetch.op.topo,
                    "num_chunks": IntegerAttr(self.num_chunks, IntegerType(64)),
                    "bounds": op.bounds,
                },
                regions=[
                    receive_chunk,
                    done_exchange,
                ],
                result_types=[op.result_types],
            )
        )

        if not prefetch.uses:
            rewriter.erase_op(prefetch.op)


class PromoteCoefficients(RewritePattern):
    """
    Promotes constant coefficients to attributes. When a `csl_stencil.access` is immediately multiplied by
    an `arith.constant` as the sole use of the accessed data, the constant is promoted to a coefficient property
    in the `csl_stencil.apply` op.
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: csl_stencil.AccessOp, rewriter: PatternRewriter, /):
        if (
            not isinstance(apply := op.get_apply(), csl_stencil.ApplyOp)
            or not op.op == apply.receive_chunk.block.args[0]
            or not isinstance(mulf := op.result.get_user_of_unique_use(), arith.MulfOp)
        ):
            return

        coeff = mulf.lhs if op.result == mulf.rhs else mulf.rhs

        if (
            not isinstance(cnst := coeff.owner, arith.ConstantOp)
            or not isinstance(dense := cnst.value, DenseIntOrFPElementsAttr)
            or not dense.is_splat()
        ):
            return

        val = dense.get_attrs()[0]
        assert isinstance(val, FloatAttr)
        apply.add_coeff(op.offset, val)
        rewriter.replace_op(mulf, [], new_results=[op.result])


class TransformPrefetch(RewritePattern):
    """
    Rewrites a prefetch into a communicate-only csl_stencil.apply
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(
        self, op: csl_stencil.PrefetchOp, rewriter: PatternRewriter, /
    ):
        a_buf = tensor.EmptyOp((), op.result.type)
        # because we are building a set of offsets, we are not retaining offset mappings
        offsets = [swap.neighbor for swap in op.swaps]

        assert isa(op.result.type, AnyTensorType)
        chunk_buf_t = TensorType(
            op.result.type.get_element_type(),
            (
                len(op.swaps),
                op.result.type.get_shape()[1] // op.num_chunks.value.data,
            ),
        )
        chunk_t = TensorType(chunk_buf_t.element_type, chunk_buf_t.get_shape()[1:])

        block = Block(arg_types=[chunk_buf_t, builtin.IndexType(), op.result.type])
        block2 = Block(arg_types=[op.input_stencil.type, op.result.type])
        block2.add_op(csl_stencil.YieldOp())

        with ImplicitBuilder(block) as (buf, offset, acc):
            dest = acc
            for i, acc_offset in enumerate(offsets):
                ac_op = csl_stencil.AccessOp(
                    buf, stencil.IndexAttr.get(*acc_offset), chunk_t
                )
                assert isa(ac_op.result.type, AnyTensorType)
                # inserts 1 (see static_sizes) 1d slice into a 2d tensor at offset (i, `offset`) (see static_offsets)
                # where the latter offset is provided dynamically (see offsets)
                dest = tensor.InsertSliceOp.get(
                    source=ac_op.result,
                    dest=dest,
                    static_sizes=[1, *ac_op.result.type.get_shape()],
                    static_offsets=[i, memref.SubviewOp.DYNAMIC_INDEX],
                    offsets=[offset],
                ).result
            csl_stencil.YieldOp(dest)

        apply_op = csl_stencil.ApplyOp(
            operands=[op.input_stencil, a_buf, [], [], []],
            regions=[Region(block), Region(block2)],
            properties={
                "swaps": op.swaps,
                "topo": op.topo,
                "num_chunks": op.num_chunks,
            },
            result_types=[[]],
        )

        rewriter.replace_matched_op([a_buf, apply_op], new_results=[a_buf.tensor])


@dataclass(frozen=True)
class ConvertStencilToCslStencilPass(ModulePass):
    name = "convert-stencil-to-csl-stencil"

    # chunks into which to slice communication
    num_chunks: int = 1

    def apply(self, ctx: Context, op: ModuleOp) -> None:
        ConvertArithToVarithPass().apply(ctx, op)
        module_pass = PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [
                    ConvertSwapToPrefetchPattern(num_chunks=self.num_chunks),
                    ConvertAccessOpPattern(),
                ]
            ),
            walk_reverse=False,
            apply_recursively=False,
        )
        module_pass.rewrite_module(op)
        PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [
                    ConvertApplyOpPattern(num_chunks=self.num_chunks),
                    PromoteCoefficients(),
                    TransformPrefetch(),
                ]
            ),
            apply_recursively=False,
            walk_reverse=True,
        ).rewrite_module(op)

        ConvertVarithToArithPass().apply(ctx, op)

        if self.num_chunks > 1:
            BackpropagateStencilShapes().apply(ctx, op)



from collections.abc import Iterable
from itertools import chain

from xdsl.context import Context
from xdsl.dialects import affine, scf
from xdsl.dialects.builtin import ModuleOp
from xdsl.ir import Operation, SSAValue
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    GreedyRewritePatternApplier,
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    op_type_rewrite_pattern,
)
from xdsl.rewriter import InsertPoint
from xdsl.traits import IsTerminator, is_side_effect_free, is_speculatable
from xdsl.transforms.common_subexpression_elimination import cse


def hoist_all(
    rewriter: PatternRewriter,
    ops: Iterable[Operation],
    at: InsertPoint,
    value_mapper: dict[SSAValue, SSAValue] | None = None,
):
    if value_mapper is None:
        value_mapper = {}
    for o in ops:
        if o.get_trait(IsTerminator):
            continue
        new_op = o.clone(value_mapper=value_mapper)
        value_mapper |= {
            old: new for old, new in zip(o.results, new_op.results, strict=True)
        }
        rewriter.insert_op(new_op, at)
        rewriter.replace_op(o, [], new_op.results)


class AffineIfHoistPattern(RewritePattern):
    """
    Hoist everything out of a pure affine.if.
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: affine.IfOp, rewriter: PatternRewriter):
        # Easy bail out for now
        if not (is_speculatable(op) and is_side_effect_free(op)):
            return

        hoist_all(
            rewriter,
            chain(op.then_region.ops, op.else_region.ops),
            InsertPoint.before(op),
        )
        if not rewriter.has_done_action:
            return
        block = op.parent
        if block:
            cse(block, rewriter)


class SCFIfHoistPattern(RewritePattern):
    """
    Hoist everything out of a pure scf.if
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: scf.IfOp, rewriter: PatternRewriter):
        # Easy bail out for now
        if not (is_speculatable(op) and is_side_effect_free(op)):
            return

        hoist_all(
            rewriter,
            chain(op.true_region.ops, op.false_region.ops),
            InsertPoint.before(op),
        )

        # Perf-friendly cleanup
        # None needed if nothing happened
        if not rewriter.has_done_action:
            return
        block = op.parent
        if block:
            # If we hoisted some ops, run CSE on that block to not keep pushing duplicates upward.
            cse(block, rewriter)


class ControlFlowHoistPass(ModulePass):
    """
    Hoist all hoistable ops from control flow ops.
    """

    name = "control-flow-hoist"

    def apply(self, ctx: Context, op: ModuleOp) -> None:
        PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [
                    AffineIfHoistPattern(),
                    SCFIfHoistPattern(),
                ]
            ),
            walk_regions_first=True,
        ).rewrite_module(op)



from collections import Counter

from xdsl.dialects.builtin import ModuleOp


def count_number_of_operations(module: ModuleOp) -> dict[str, int]:
    """
    This function takes a ModuleOp and returns a dictionary containing the number of
    occurences of each Operation in the ModuleOp.
    """
    return Counter(op.name for op in module.walk())


def get_diff_operation_count(
    input_operation_count_tuple: tuple[tuple[str, int], ...],
    output_operation_count_tuple: tuple[tuple[str, int], ...],
) -> tuple[tuple[str, int, str], ...]:
    """
    Function returning a tuple of tuples containing the diff of the input and output
    operation name and count.
    """
    input_op_count_dict = dict(input_operation_count_tuple)
    output_op_count_dict = dict(output_operation_count_tuple)
    all_keys = {*input_op_count_dict, *output_op_count_dict}

    res: dict[str, tuple[int, str]] = {}
    for k in all_keys:
        input_count = input_op_count_dict.get(k, 0)
        output_count = output_op_count_dict.get(k, 0)
        diff = output_count - input_count

        # convert diff to string
        if diff == 0:
            diff_str = "="
        elif diff > 0:
            diff_str = f"+{diff}"
        else:
            diff_str = str(diff)

        res[k] = (output_count, diff_str)

    return tuple((k, v0, v1) for (k, (v0, v1)) in sorted(res.items()))



from xdsl.dialects import arith
from xdsl.dialects.builtin import (
    AffineMapAttr,
    IntegerAttr,
)
from xdsl.dialects.csl import csl
from xdsl.ir import OpResult
from xdsl.ir.affine import AffineMap
from xdsl.pattern_rewriter import (
    PatternRewriter,
    RewritePattern,
    op_type_rewrite_pattern,
)
from xdsl.utils.hints import isa


class GetDsdAndOffsetFolding(RewritePattern):
    """
    Folds a `csl.get_mem_dsd` immediately followed by a `csl.increment_dsd_offset`
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: csl.GetMemDsdOp, rewriter: PatternRewriter) -> None:
        # single use that is `@increment_dsd_offset`
        if not isinstance(
            offset_op := op.result.get_user_of_unique_use(), csl.IncrementDsdOffsetOp
        ):
            return
        # only works on 1d
        if len(op.sizes) > 1:
            return

        # check if we can promote arith.const to property
        if (
            isinstance(offset_op.offset, OpResult)
            and isinstance(cnst := offset_op.offset.op, arith.ConstantOp)
            and isa(attr_val := cnst.value, IntegerAttr)
        ):
            tensor_access = AffineMap.from_callable(
                lambda x: (x + attr_val.value.data,)
            )
            if op.tensor_access:
                tensor_access = tensor_access.compose(op.tensor_access.data)
            rewriter.replace_matched_op(
                new_op := csl.GetMemDsdOp.build(
                    operands=[op.base_addr, op.sizes],
                    result_types=op.result_types,
                    properties={
                        **op.properties,
                        "tensor_access": AffineMapAttr(tensor_access),
                    },
                )
            )
            rewriter.replace_op(offset_op, [], new_results=[new_op.result])


class GetDsdAndLengthFolding(RewritePattern):
    """
    Folds a `csl.get_mem_dsd` immediately followed by a `csl.set_dsd_length`
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: csl.GetMemDsdOp, rewriter: PatternRewriter) -> None:
        # single use that is `@set_dsd_length`
        if not isinstance(
            size_op := op.result.get_user_of_unique_use(), csl.SetDsdLengthOp
        ):
            return
        # only works on 1d
        if len(op.sizes) > 1:
            return

        rewriter.replace_op(
            size_op,
            csl.GetMemDsdOp.build(
                operands=[op.base_addr, [size_op.length]],
                result_types=op.result_types,
                properties=op.properties.copy(),
            ),
        )
        rewriter.erase_matched_op()


class GetDsdAndStrideFolding(RewritePattern):
    """
    Folds a `csl.get_mem_dsd` immediately followed by a `csl.set_dsd_stride`
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: csl.GetMemDsdOp, rewriter: PatternRewriter) -> None:
        # single use that is `@set_dsd_stride`
        if not isinstance(
            stride_op := op.result.get_user_of_unique_use(), csl.SetDsdStrideOp
        ):
            return
        # only works on 1d and default (unspecified) tensor_access
        if len(op.sizes) > 1 or op.tensor_access:
            return

        # check if we can promote arith.const to property
        if (
            isinstance(stride_op.stride, OpResult)
            and isinstance(cnst := stride_op.stride.op, arith.ConstantOp)
            and isa(attr_val := cnst.value, IntegerAttr)
        ):
            tensor_access = AffineMap.from_callable(
                lambda x: (x * attr_val.value.data,)
            )
            rewriter.replace_matched_op(
                new_op := csl.GetMemDsdOp.build(
                    operands=[op.base_addr, op.sizes],
                    result_types=op.result_types,
                    properties={
                        **op.properties,
                        "tensor_access": AffineMapAttr(tensor_access),
                    },
                )
            )
            rewriter.replace_op(stride_op, [], new_results=[new_op.result])


class ChainedDsdOffsetFolding(RewritePattern):
    """
    Folds a chain of `csl.increment_dsd_offset`
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(
        self, op: csl.IncrementDsdOffsetOp, rewriter: PatternRewriter
    ) -> None:
        # single use that is `@increment_dsd_offset`
        if not isinstance(
            next_op := op.result.get_user_of_unique_use(), csl.IncrementDsdOffsetOp
        ):
            return

        # check if we can promote arith.const to property
        if op.elem_type == next_op.elem_type:
            rewriter.replace_op(
                next_op,
                [
                    new_offset := arith.AddiOp(op.offset, next_op.offset),
                    csl.IncrementDsdOffsetOp(
                        operands=[op.op, new_offset.result],
                        properties=op.properties.copy(),
                        result_types=op.result_types,
                    ),
                ],
            )
            rewriter.erase_matched_op()


class ChainedDsdLengthFolding(RewritePattern):
    """
    Folds a chain of `csl.set_dsd_length`
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(
        self, op: csl.SetDsdLengthOp, rewriter: PatternRewriter
    ) -> None:
        # single use that is `@set_dsd_length`
        if not isinstance(
            next_op := op.result.get_user_of_unique_use(), csl.SetDsdLengthOp
        ):
            return

        # check if we can promote arith.const to property
        rewriter.replace_matched_op(
            rebuilt := csl.SetDsdLengthOp(
                operands=[op.op, next_op.length],
                properties=op.properties.copy(),
                result_types=op.result_types,
            ),
        )
        rewriter.replace_op(next_op, [], new_results=[rebuilt.result])


class ChainedDsdStrideFolding(RewritePattern):
    """
    Folds a chain of `csl.set_dsd_stride`
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(
        self, op: csl.SetDsdStrideOp, rewriter: PatternRewriter
    ) -> None:
        # single use that is `@set_dsd_stride`
        if not isinstance(
            next_op := op.result.get_user_of_unique_use(), csl.SetDsdStrideOp
        ):
            return

        # check if we can promote arith.const to property
        rewriter.replace_matched_op(
            rebuilt := csl.SetDsdStrideOp(
                operands=[op.op, next_op.stride],
                properties=op.properties.copy(),
                result_types=op.result_types,
            )
        )
        rewriter.replace_op(next_op, [], new_results=[rebuilt.result])



from xdsl.dialects.test import TestOp
from xdsl.ir import AttributeCovT, OpResult


def create_ssa_value(t: AttributeCovT) -> OpResult[AttributeCovT]:
    op = TestOp(result_types=(t,))
    return op.results[0]  # pyright: ignore[reportReturnType]



import pytest

from xdsl.context import Context
from xdsl.dialects import test
from xdsl.dialects.arith import AddiOp, Arith, ConstantOp, SubiOp
from xdsl.dialects.builtin import (
    Builtin,
    Float64Type,
    IndexType,
    IntegerAttr,
    IntegerType,
    ModuleOp,
    StringAttr,
    TensorType,
    i32,
    i64,
)
from xdsl.dialects.cf import Cf
from xdsl.dialects.func import Func
from xdsl.ir import (
    Block,
    ErasedSSAValue,
    Operation,
    Region,
    SSAValue,
)
from xdsl.irdl import (
    IRDLOperation,
    irdl_op_definition,
    operand_def,
    prop_def,
    var_region_def,
)
from xdsl.parser import Parser
from xdsl.utils.exceptions import VerifyException
from xdsl.utils.test_value import create_ssa_value


@irdl_op_definition
class TestWithPropOp(IRDLOperation):
    name = "test.op_with_prop"

    prop = prop_def()


def test_ops_accessor():
    a = ConstantOp.from_int_and_width(1, i32)
    b = ConstantOp.from_int_and_width(2, i32)
    # Operation to add these constants
    c = AddiOp(a, b)

    block0 = Block([a, b, c])
    # Create a region to include a, b, c
    region = Region(block0)

    assert len(region.ops) == 3
    assert len(region.block.ops) == 3

    # Operation to subtract b from a
    d = SubiOp(a, b)

    assert d.results[0] != c.results[0]

    assert c.lhs.owner is a
    assert c.rhs.owner is b
    assert d.lhs.owner is a
    assert d.rhs.owner is b


def test_ops_accessor_II():
    a = ConstantOp.from_int_and_width(1, i32)
    b = ConstantOp.from_int_and_width(2, i32)
    # Operation to add these constants
    c = AddiOp(a, b)

    block0 = Block([a, b, c])
    # Create a region to include a, b, c
    region = Region(block0)

    assert len(region.ops) == 3
    assert len(region.block.ops) == 3

    # Operation to subtract b from a
    d = SubiOp(a, b)

    assert d.results[0] != c.results[0]

    # Erase operations and block
    region2 = Region()
    region.move_blocks(region2)

    region2.block.erase_op(a, safe_erase=False)
    region2.block.erase_op(b, safe_erase=False)
    region2.block.erase_op(c, safe_erase=False)

    assert isinstance(c.lhs, ErasedSSAValue)
    assert isinstance(c.rhs, ErasedSSAValue)
    assert c.lhs.owner is a
    assert c.rhs.owner is b

    region2.detach_block(block0)
    region2.drop_all_references()

    assert len(region2.blocks) == 0


def test_ops_accessor_III():
    # Create constants and add them, add them in blocks, blocks in
    # a region and create a function
    a = ConstantOp(IntegerAttr.from_int_and_width(1, 32), i32)
    b = ConstantOp(IntegerAttr.from_int_and_width(2, 32), i32)
    c = ConstantOp(IntegerAttr.from_int_and_width(3, 32), i32)
    d = ConstantOp(IntegerAttr.from_int_and_width(4, 32), i32)

    # Operation to add these constants
    e = AddiOp(a, b)
    f = AddiOp(c, d)

    # Create Blocks and Regions
    block0 = Block([a, b, e])
    block1 = Block([c, d, f])
    block2 = Block()

    region0 = Region([block0, block1])
    region1 = Region(block2)

    with pytest.raises(
        ValueError,
        match="'ops' property of Region class is only available for single-block regions.",
    ):
        region0.ops

    with pytest.raises(
        ValueError,
        match="'op' property of Region class is only available for single-operation single-block regions.",
    ):
        region0.op

    with pytest.raises(
        ValueError,
        match="'op' property of Region class is only available for single-operation single-block regions.",
    ):
        region1.op

    with pytest.raises(
        ValueError,
        match="Block is not a child of the region.",
    ):
        region1.detach_block(block0)

    region0.detach_block(block0)
    region0.detach_block(0)
    with pytest.raises(IndexError):
        region0.detach_block(1)


def test_op_operands_assign():
    """Test that we can directly assign `op.operands`."""
    val1, val2 = create_ssa_value(i32), create_ssa_value(i32)
    op = test.TestOp.create(operands=[val1, val2])
    op.operands = [val2, val1]
    op.verify()

    assert val1.has_one_use()
    assert val2.has_one_use()
    assert tuple(op.operands) == (val2, val1)


def test_op_operands_indexing():
    """Test `__getitem__`, `__setitem__`, and `__len__` on `op.operands`."""
    val1, val2 = create_ssa_value(i32), create_ssa_value(i32)
    op = test.TestOp.create(operands=[val1, val2])
    op.verify()

    assert op.operands[0] == val1
    assert op.operands[1] == val2
    assert op.operands[-1] == val2
    assert op.operands[0:2] == (val1, val2)

    op.operands[0] = val2
    op.verify()

    assert not list(val1.uses)
    assert len(list(val2.uses)) == 2
    assert tuple(op.operands) == (val2, val2)


def test_op_operands_comparison():
    """Test `__eq__`, and `__hash__` on `op.operands`."""
    val1, val2 = create_ssa_value(i32), create_ssa_value(i32)
    op1 = test.TestOp.create(operands=[val1, val2])
    op2 = test.TestOp.create(operands=[val1, val2])
    op1.verify()
    op2.verify()

    assert op1.operands == op2.operands
    assert hash(op1.operands) == hash(op2.operands)

    op1.operands[0] = val2
    op1.verify()

    assert op1.operands != op2.operands


def test_op_clone():
    a = TestWithPropOp.create(
        properties={"prop": i32}, attributes={"attr": i64}, result_types=(i32,)
    )
    a.results[0].name_hint = "name_hint"
    b = a.clone()

    assert a is not b
    assert a.is_structurally_equivalent(b)

    # Name hints

    c = a.clone(clone_name_hints=True)
    d = a.clone(clone_name_hints=False)
    assert a is not c
    assert a is not d

    assert a.is_structurally_equivalent(c)
    assert a.is_structurally_equivalent(d)

    assert b.results[0].name_hint == "name_hint"
    assert c.results[0].name_hint == "name_hint"
    assert d.results[0].name_hint is None


def test_op_clone_with_regions():
    # Children
    ca0 = test.TestOp.create(result_types=(i32,))
    ca0.results[0].name_hint = "a"
    ca1 = test.TestOp.create(result_types=(i32,))
    ca1.results[0].name_hint = "b"
    # Parent
    pa = test.TestOp.create(
        regions=[
            Region([Block([ca0], arg_types=(i32,))]),
            Region([Block([ca1], arg_types=(i32,))]),
        ]
    )
    pa.regions[0].block.args[0].name_hint = "ca0"
    pa.regions[1].block.args[0].name_hint = "ca1"

    pb = pa.clone()
    assert pa is not pb

    assert len(pb.regions[0].ops) == 1
    assert len(pb.regions[1].ops) == 1

    for op0_region, cloned_op_region in zip(pa.regions, pb.regions):
        for op0_region_op, cloned_region_op in zip(
            op0_region.ops, cloned_op_region.ops
        ):
            assert op0_region_op is not cloned_region_op

    pc = pa.clone(clone_name_hints=True)
    pd = pa.clone(clone_name_hints=False)

    def name_hints(op: Operation):
        for o in op.walk():
            for res in o.results:
                yield res.name_hint
            for r in o.regions:
                for b in r.blocks:
                    for arg in b.args:
                        yield arg.name_hint

    assert tuple(name_hints(pa)) == ("ca0", "ca1", "a", "b")
    assert tuple(name_hints(pb)) == ("ca0", "ca1", "a", "b")
    assert tuple(name_hints(pc)) == ("ca0", "ca1", "a", "b")
    assert tuple(name_hints(pd)) == (None, None, None, None)


def test_block_branching_to_another_region_wrong():
    """
    Tests that an operation cannot have successors that branch to blocks of
    another region.
    """

    block1 = Block([test.TestOp.create(), test.TestOp.create()])
    region1 = Region([block1])

    op0 = test.TestOp.create(successors=[block1])
    block0 = Block([op0])
    region0 = Region([block0])
    region0 = test.TestOp.create(regions=[region0, region1])

    outer_block = Block([region0])

    with pytest.raises(
        VerifyException,
        match="is branching to a block of a different region",
    ):
        outer_block.verify()


def test_block_not_branching_to_another_region():
    """
    Tests that an operation can have successors that branch to blocks of the
    same region.
    """
    block0 = Block()

    op0 = test.TestTermOp.create(successors=[block0])
    block1 = Block([op0])

    region0 = Region([block0, block1])

    region0.verify()


def test_empty_block_with_no_parent_region_requires_no_terminator():
    """
    Tests that an empty block belonging no parent region requires no terminator
    operation.
    """
    block0 = Block([])

    block0.verify()


def test_empty_block_with_orphan_single_block_parent_region_requires_no_terminator():
    """
    Tests that an empty block belonging to a single-block region with no parent
    operation requires no terminator operation.
    """
    block0 = Block([])
    region0 = Region([block0])

    region0.verify()


def test_empty_block_with_single_block_parent_region_requires_terminator():
    """
    Tests that an empty block belonging to a single-block region in a parent
    operation requires terminator operation.
    """
    block0 = Block([])
    region0 = Region([block0])
    op0 = test.TestOp.create(regions=[region0])

    with pytest.raises(
        VerifyException,
        match="contains empty block in single-block region that expects at least a terminator",
    ):
        op0.verify()


def test_split_block_first():
    old_block = Block((test.TestOp(), test.TestOp(), test.TestOp()))
    region = Region(old_block)
    a, b, c = old_block.ops

    # Check preconditions

    assert old_block.parent is region
    assert list(region.blocks) == [old_block]

    assert old_block.first_op is a
    assert old_block.last_op is c

    assert a.parent is old_block
    assert b.parent is old_block
    assert c.parent is old_block

    assert a.next_op is b
    assert b.next_op is c
    assert c.next_op is None

    assert a.prev_op is None
    assert b.prev_op is a
    assert c.prev_op is b

    new_block = old_block.split_before(a)

    # Check postconditions

    assert old_block.parent is region
    assert new_block.parent is region
    assert list(region.blocks) == [old_block, new_block]

    assert old_block.first_op is None
    assert old_block.last_op is None
    assert new_block.first_op is a
    assert new_block.last_op is c

    assert a.parent is new_block
    assert b.parent is new_block
    assert c.parent is new_block

    assert a.next_op is b
    assert b.next_op is c
    assert c.next_op is None

    assert a.prev_op is None
    assert b.prev_op is a
    assert c.prev_op is b


def test_split_block_middle():
    old_block = Block((test.TestOp(), test.TestOp(), test.TestOp()))
    region = Region(old_block)
    a, b, c = old_block.ops

    # Check preconditions

    assert old_block.parent is region
    assert list(region.blocks) == [old_block]

    assert old_block.first_op is a
    assert old_block.last_op is c

    assert a.parent is old_block
    assert b.parent is old_block
    assert c.parent is old_block

    assert a.next_op is b
    assert b.next_op is c
    assert c.next_op is None

    assert a.prev_op is None
    assert b.prev_op is a
    assert c.prev_op is b

    new_block = old_block.split_before(b)

    # Check postconditions

    assert old_block.parent is region
    assert new_block.parent is region
    assert list(region.blocks) == [old_block, new_block]

    assert old_block.first_op is a
    assert old_block.last_op is a
    assert new_block.first_op is b
    assert new_block.last_op is c

    assert a.parent is old_block
    assert b.parent is new_block
    assert c.parent is new_block

    assert a.next_op is None
    assert b.next_op is c
    assert c.next_op is None

    assert a.prev_op is None
    assert b.prev_op is None
    assert c.prev_op is b


def test_split_block_last():
    old_block = Block((test.TestOp(), test.TestOp(), test.TestOp()))
    region = Region(old_block)
    a, b, c = old_block.ops

    # Check preconditions

    assert old_block.parent is region
    assert list(region.blocks) == [old_block]

    assert old_block.first_op is a
    assert old_block.last_op is c

    assert a.parent is old_block
    assert b.parent is old_block
    assert c.parent is old_block

    assert a.next_op is b
    assert b.next_op is c
    assert c.next_op is None

    assert a.prev_op is None
    assert b.prev_op is a
    assert c.prev_op is b

    new_block = old_block.split_before(c)

    # Check postconditions

    assert old_block.parent is region
    assert new_block.parent is region
    assert list(region.blocks) == [old_block, new_block]

    assert old_block.first_op is a
    assert old_block.last_op is b
    assert new_block.first_op is c
    assert new_block.last_op is c

    assert a.parent is old_block
    assert b.parent is old_block
    assert c.parent is new_block

    assert a.next_op is b
    assert b.next_op is None
    assert c.next_op is None

    assert a.prev_op is None
    assert b.prev_op is a
    assert c.prev_op is None


def test_split_block_args():
    old_block = Block((test.TestOp(), test.TestOp(), test.TestOp()))
    region = Region(old_block)
    _, op, _ = region.block.ops

    new_block = old_block.split_before(op, arg_types=(i32, i64))

    arg_types = new_block.arg_types
    assert arg_types == (i32, i64)


def test_region_clone_into_circular_blocks():
    """
    Test that cloning a region with circular block dependency works.
    """
    region_str = """
    {
    ^0:
        "test.op"() [^1] : () -> ()
    ^1:
        "test.op"() [^0] : () -> ()
    }
    """
    ctx = Context(allow_unregistered=True)
    region = Parser(ctx, region_str).parse_region()

    region2 = Region()
    region.clone_into(region2)

    assert region.is_structurally_equivalent(region2)


def test_op_with_successors_not_in_block():
    block0 = Block()
    op0 = test.TestTermOp.create(successors=[block0])

    with pytest.raises(
        VerifyException,
        match="with block successors does not belong to a block or a region",
    ):
        op0.verify()


def test_op_with_successors_not_in_region():
    block1 = Block()

    op0 = test.TestOp.create(successors=[block1])
    block0 = Block([op0])

    with pytest.raises(
        VerifyException,
        match="with block successors does not belong to a block or a region",
    ):
        block0.verify()


def test_non_empty_block_with_single_block_parent_region_must_have_terminator():
    """
    Tests that an non-empty block belonging to a single-block region with parent
    operation cannot have an operation that is not a terminator.
    """
    block1 = Block([test.TestOp.create()])
    region0 = Region([block1])
    op0 = test.TestOp.create(regions=[region0])

    with pytest.raises(
        VerifyException,
        match="terminates block in single-block region but is not a terminator",
    ):
        op0.verify()


def test_non_empty_block_with_single_block_parent_region_with_terminator():
    """
    Tests that an non-empty block belonging to a single-block region with parent
    operation must have at least a terminator operation.
    """
    block0 = Block([test.TestTermOp.create()])
    region0 = Region([block0])
    op0 = test.TestOp.create(regions=[region0])

    op0.verify()


def test_non_empty_block_with_parent_region_can_have_terminator_with_successors():
    """
    Tests that an non-empty block belonging to a multi-block region with parent
    operation requires terminator operation.
    The terminator operation may have successors.
    """
    block0 = Block()
    block1 = Block([test.TestTermOp.create(successors=[block0])])
    region0 = Region([block0, block1])
    op0 = test.TestOp.create(regions=[region0])

    op0.verify()


def test_non_empty_block_with_parent_region_requires_terminator_without_successors():
    """
    Tests that an non-empty block belonging to a multi-block region with parent
    operation requires terminator operation.
    The terminator operation may not have successors.
    """
    block0 = Block()
    block1 = Block([test.TestOp.create()])
    region0 = Region([block0, block1])
    op0 = test.TestOp.create(regions=[region0])

    with pytest.raises(
        VerifyException,
        match="terminates block in multi-block region but is not a terminator",
    ):
        op0.verify()


def test_non_empty_block_with_parent_region_requires_terminator_with_successors():
    """
    Tests that an non-empty block belonging to a multi-block region with parent
    operation requires terminator operation.
    The terminator operation may have successors.
    """
    block0 = Block()

    op0 = test.TestOp.create(successors=[block0])
    block1 = Block([op0])

    region0 = Region([block0, block1])

    with pytest.raises(
        VerifyException,
        match="terminates block in multi-block region but is not a terminator",
    ):
        region0.verify()


def test_non_empty_block_with_parent_region_has_successors_but_not_last_block_op():
    """
    Tests that an non-empty block belonging to a multi-block region with parent
    operation requires terminator operation.
    """
    block0 = Block()
    block1 = Block([test.TestOp.create(successors=[block0]), test.TestOp.create()])
    region0 = Region([block0, block1])
    op0 = test.TestOp.create(regions=[region0])

    with pytest.raises(
        VerifyException,
        match="with block successors must terminate its parent block",
    ):
        op0.verify()


##################### Testing is_structurally_equal #####################

program_region = """
"builtin.module"() ({
  %0 = "arith.constant"() {"value" = 1 : i32} : () -> i32
}) : () -> ()
"""

program_region_2 = """
"builtin.module"() ({
  %0 = "arith.constant"() {"value" = 2 : i32} : () -> i32
}) : () -> ()
"""

program_region_2_diff_name = """
"builtin.module"() ({
  %cst = "arith.constant"() {"value" = 2 : i32} : () -> i32
}) : () -> ()
"""

program_region_2_diff_type = """
"builtin.module"() ({
  %0 = "arith.constant"() {"value" = 2 : i64} : () -> i64
}) : () -> ()
"""

program_add = """
"builtin.module"() ({
  %0 = "arith.constant"() {"value" = 1 : i32} : () -> i32
  %1 = "arith.constant"() {"value" = 2 : i32} : () -> i32
  %2 = "arith.addi"(%0, %1) : (i32, i32) -> i32
}) : () -> ()
"""

program_add_2 = """
"builtin.module"() ({
  %0 = "arith.constant"() {"value" = 1 : i32} : () -> i32
  %1 = "arith.constant"() {"value" = 2 : i32} : () -> i32
  %2 = "arith.addi"(%1, %0) : (i32, i32) -> i32
}) : () -> ()
"""

program_func = """
"builtin.module"() ({
  "func.func"() ({
  ^0(%0 : i32, %1 : i32):
    %2 = "arith.addi"(%0, %1) : (i32, i32) -> i32
    "func.return"(%2) : (i32) -> ()
  }) {"sym_name" = "test", "function_type" = (i32, i32) -> i32, "sym_visibility" = "private"} : () -> ()
}) : () -> ()
"""

program_successors = """
"builtin.module"() ({
  "func.func"() ({
  ^0:
    "cf.br"() [^1] : () -> ()
  ^1:
    "cf.br"() [^0] : () -> ()
  }) {"sym_name" = "unconditional_br", "function_type" = () -> (), "sym_visibility" = "private"} : () -> ()
}) : () -> ()
"""


@pytest.mark.parametrize(
    "args, expected_result",
    [
        ([program_region, program_region], True),
        ([program_region_2, program_region_2], True),
        ([program_region_2_diff_type, program_region_2_diff_type], True),
        ([program_region_2_diff_name, program_region_2_diff_name], True),
        ([program_region, program_region_2], False),
        ([program_region_2, program_region_2_diff_type], False),
        ([program_region_2, program_region_2_diff_name], True),
        ([program_add, program_add], True),
        ([program_add_2, program_add_2], True),
        ([program_add, program_add_2], False),
        ([program_func, program_func], True),
        ([program_successors, program_successors], True),
        ([program_successors, program_func], False),
        ([program_successors, program_add], False),
    ],
)
def test_is_structurally_equivalent(args: list[str], expected_result: bool):
    ctx = Context()
    ctx.load_dialect(Builtin)
    ctx.load_dialect(Func)
    ctx.load_dialect(Arith)
    ctx.load_dialect(Cf)

    parser = Parser(ctx, args[0])
    lhs: Operation = parser.parse_op()

    parser = Parser(ctx, args[1])
    rhs: Operation = parser.parse_op()

    assert lhs.is_structurally_equivalent(rhs) == expected_result


def test_is_structurally_equivalent_properties():
    op32 = TestWithPropOp.create(properties={"prop": i32})
    op32prime = TestWithPropOp.create(properties={"prop": i32})
    op64 = TestWithPropOp.create(properties={"prop": i64})
    assert op32.is_structurally_equivalent(op32prime)
    assert not op32.is_structurally_equivalent(op64)


def test_is_structurally_equivalent_free_operands():
    val1 = create_ssa_value(i32)
    val2 = create_ssa_value(i64)
    op1 = test.TestOp.create(operands=[val1, val2])
    op2 = test.TestOp.create(operands=[val1, val2])
    assert op1.is_structurally_equivalent(op2)


def test_is_structurally_equivalent_free_operands_fail():
    val1 = create_ssa_value(i32)
    val2 = create_ssa_value(i32)
    op1 = test.TestOp.create(operands=[val1])
    op2 = test.TestOp.create(operands=[val2])
    assert not op1.is_structurally_equivalent(op2)


def test_is_structurally_equivalent_incompatible_ir_nodes():
    program_func = """
"builtin.module"() ({
  "func.func"() ({
  ^0(%0 : i32, %1 : i32):
    %2 = "arith.addi"(%0, %1) : (i32, i32) -> i32
    %3 = "arith.constant"() {"value" = 2 : i32} : () -> i32
    "func.return"(%3) : (i32) -> ()
  ^1(%4 : i32, %5 : i32):
    "func.return"(%4) : (i32) -> ()
  }) {"sym_name" = "test", "function_type" = (i32, i32) -> i32, "sym_visibility" = "private"} : () -> ()
}) : () -> ()
"""
    ctx = Context()
    ctx.load_dialect(Builtin)
    ctx.load_dialect(Func)
    ctx.load_dialect(Arith)
    ctx.load_dialect(Cf)

    parser = Parser(ctx, program_func)
    program = parser.parse_operation()

    assert isinstance(program, ModuleOp)

    assert program.is_structurally_equivalent(program.regions[0]) is False
    assert program.is_structurally_equivalent(program.regions[0].blocks[0]) is False
    assert program.regions[0].is_structurally_equivalent(program) is False
    assert program.regions[0].blocks[0].is_structurally_equivalent(program) is False

    func_op = program.ops.first
    assert func_op is not None

    block = func_op.regions[0].blocks[0]
    ops = list(block.ops)
    assert not ops[0].is_structurally_equivalent(ops[1])
    assert not block.is_structurally_equivalent(func_op.regions[0].blocks[1])


def test_descriptions():
    a = ConstantOp.from_int_and_width(1, 32)

    assert str(a.value) == "1 : i32"
    assert f"{a.value}" == "1 : i32"

    assert str(a) == "%0 = arith.constant 1 : i32"
    assert f"{a}" == "ConstantOp(%0 = arith.constant 1 : i32)"

    m = ModuleOp([a])

    assert (
        str(m)
        == """\
builtin.module {
  %0 = arith.constant 1 : i32
}"""
    )

    assert (
        f"{m}"
        == """\
ModuleOp(
\tbuiltin.module {
\t  %0 = arith.constant 1 : i32
\t}
)"""
    )


# ToDo: Create this op without IRDL itself, since it tests fine grained
# stuff which is supposed to be used with IRDL or PDL.
@irdl_op_definition
class MultipleRegionsOp(IRDLOperation):
    name = "test.custom_op_with_multiple_regions"
    region = var_region_def()


def test_region_index_fetch():
    a = ConstantOp.from_int_and_width(1, 32)
    b = ConstantOp.from_int_and_width(2, 32)
    c = ConstantOp.from_int_and_width(3, 32)
    d = ConstantOp.from_int_and_width(4, 32)

    region0 = Region([Block([a])])
    region1 = Region([Block([b])])
    region2 = Region([Block([c])])
    region3 = Region([Block([d])])

    op = MultipleRegionsOp.build(regions=[[region0, region1, region2, region3]])

    assert op.get_region_index(region0) == 0
    assert op.get_region_index(region1) == 1
    assert op.get_region_index(region2) == 2
    assert op.get_region_index(region3) == 3


def test_region_index_fetch_region_unavailability():
    a = ConstantOp.from_int_and_width(1, 32)
    b = ConstantOp.from_int_and_width(2, 32)

    region0 = Region([Block([a])])
    region1 = Region([Block([b])])

    op = MultipleRegionsOp.build(regions=[[region0]])

    assert op.get_region_index(region0) == 0
    with pytest.raises(ValueError, match="Region is not attached to the operation."):
        op.get_region_index(region1)


def test_detach_region():
    a = ConstantOp.from_int_and_width(1, 32)
    b = ConstantOp.from_int_and_width(2, 32)
    c = ConstantOp.from_int_and_width(3, 32)

    region0 = Region([Block([a])])
    region1 = Region([Block([b])])
    region2 = Region([Block([c])])

    op = MultipleRegionsOp.build(regions=[[region0, region1, region2]])

    assert op.detach_region(1) == region1
    assert op.detach_region(region0) == region0
    assert len(op.regions) == 1
    assert op.get_region_index(region2) == 0


def test_region_hashable():
    a = Region()
    b = Region()
    assert a == a
    assert a != b
    assert hash(a) == hash(a)
    assert a in {a}
    assert b not in {a}


@irdl_op_definition
class CustomVerifyOp(IRDLOperation):
    name = "test.custom_verify_op"
    val = operand_def(i64)

    @staticmethod
    def get(val: SSAValue):
        return CustomVerifyOp.build(operands=[val])

    def verify_(self):
        raise Exception("Custom Verification Check")


def test_op_custom_verify_is_called():
    a = ConstantOp.from_int_and_width(1, i64)
    b = CustomVerifyOp.get(a.result)
    with pytest.raises(Exception, match="Custom Verification Check"):
        b.verify()


def test_op_custom_verify_is_done_last():
    a = ConstantOp.from_int_and_width(1, i32)
    # CustomVerify expects a i64, not i32
    b = CustomVerifyOp.get(a.result)
    with pytest.raises(
        VerifyException,
        match="operand at position 0 does not verify:\nExpected attribute i64 but got i32",
    ):
        b.verify()


def test_block_walk():
    a = ConstantOp.from_int_and_width(1, 32)
    b = ConstantOp.from_int_and_width(2, 32)
    c = ConstantOp.from_int_and_width(3, 32)

    ops = [a, b, c]
    block = Block(ops)

    assert list(block.walk()) == [a, b, c]
    assert list(block.walk(reverse=True)) == [c, b, a]


def test_region_walk():
    a = ConstantOp.from_int_and_width(1, 32)
    b = ConstantOp.from_int_and_width(2, 32)

    block_a = Block([a])
    block_b = Block([b])

    region = Region([block_a, block_b])

    assert list(region.walk()) == [a, b]
    assert list(region.walk(reverse=True)) == [b, a]


def test_op_walk():
    a = ConstantOp.from_int_and_width(1, 32)
    b = ConstantOp.from_int_and_width(2, 32)

    block_a = Block([a])
    block_b = Block([b])

    region_a = Region(block_a)
    region_b = Region(block_b)

    op_multi_region = test.TestOp.create(regions=[region_a, region_b])

    assert list(op_multi_region.walk()) == [op_multi_region, a, b]
    assert list(op_multi_region.walk(reverse=True)) == [op_multi_region, b, a]
    assert list(op_multi_region.walk(region_first=True)) == [a, b, op_multi_region]
    assert list(op_multi_region.walk(region_first=True, reverse=True)) == [
        b,
        a,
        op_multi_region,
    ]


def test_region_clone():
    a = ConstantOp.from_int_and_width(1, 32)
    block_a = Block([a])
    region = Region(block_a)
    region2 = region.clone()
    assert region.is_structurally_equivalent(region2)


def test_get_attr_or_prop():
    a = test.TestOp.create(
        attributes={"attr": StringAttr("attr"), "attr_and_prop": StringAttr("attr")},
        properties={"prop": StringAttr("prop"), "attr_and_prop": StringAttr("prop")},
    )
    assert a.get_attr_or_prop("attr") == StringAttr("attr")
    assert a.get_attr_or_prop("prop") == StringAttr("prop")
    assert a.get_attr_or_prop("attr_and_prop") == StringAttr("prop")
    assert a.get_attr_or_prop("none") is None


def test_dialect_name():
    class MyOperation(Operation):
        name = "dialect.op"

    assert MyOperation.dialect_name() == "dialect"


def test_replace_by_if():
    a = create_ssa_value(i32)
    b = test.TestOp((a,))
    c = test.TestOp((a,))

    assert set(u.operation for u in a.uses) == {b, c}

    d = create_ssa_value(i32)
    a.replace_by_if(d, lambda u: u.operation is not c)

    assert set(u.operation for u in a.uses) == {c}
    assert set(u.operation for u in d.uses) == {b}


def test_same_block():
    op1 = test.TestOp()
    op2 = test.TestOp()
    test.TestOp(regions=[Region(Block([op1, op2]))])

    assert op1.is_before_in_block(op2)
    assert not op2.is_before_in_block(op1)
    assert not op1.is_before_in_block(op1)


def test_different_blocks():
    op1 = test.TestOp()
    op2 = test.TestOp()
    test.TestOp(regions=[Region(Block([op1])), Region(Block([op2]))])

    assert not op1.is_before_in_block(op2)
    assert not op2.is_before_in_block(op1)


def test_ancestor_block_in_region_nested():
    # lvl 3
    op_3 = test.TestOp()

    # lvl 2
    blk_2 = Block([op_3])
    op_2 = test.TestOp(regions=[Region(blk_2)])

    # lvl 1
    blk_1 = Block([op_2])
    reg_1 = Region(blk_1)
    test.TestOp(regions=[reg_1])

    assert reg_1.find_ancestor_block_in_region(blk_2) is blk_1
    assert reg_1.find_ancestor_block_in_region(blk_1) is blk_1


def test_ancestor_block_in_region_different_region():
    # lvl 3
    op_3 = test.TestOp()

    # lvl 2
    blk_2 = Block([op_3])
    op_2 = test.TestOp(regions=[Region(blk_2)])

    # lvl 1
    blk_1 = Block([op_2])
    reg_1_1 = Region(blk_1)
    reg_1_2 = Region()
    test.TestOp(regions=[reg_1_1, reg_1_2])

    assert reg_1_2.find_ancestor_block_in_region(blk_2) is None


def test_find_ancestor_op_in_block():
    op1 = test.TestOp()
    op2 = test.TestOp(regions=[Region(Block([op1]))])
    op3 = test.TestOp(regions=[Region(Block([op2]))])

    blk_top = Block([op3])
    op4 = test.TestOp(regions=[Region(blk_top)])

    assert blk_top.find_ancestor_op_in_block(op1) is op3
    assert blk_top.find_ancestor_op_in_block(op4) is None


def test_ssa_get_on_ssa():
    ssa_value = create_ssa_value(i32)

    assert SSAValue.get(ssa_value) == ssa_value
    assert SSAValue.get(ssa_value, type=IntegerType) == ssa_value
    assert SSAValue.get(ssa_value, type=IntegerType).type == i32

    with pytest.raises(
        ValueError,
        match="SSAValue.get: Expected <class 'xdsl.dialects.builtin.IndexType'> but got SSAValue with type i32",
    ):
        SSAValue.get(ssa_value, type=IndexType)


def test_ssa_get_with_typeform():
    ssa_value = create_ssa_value(i32)

    assert SSAValue.get(ssa_value, type=IntegerType | IndexType) == ssa_value
    assert SSAValue.get(ssa_value, type=IntegerType | IndexType).type == i32

    with pytest.raises(
        ValueError,
        match="SSAValue.get: Expected xdsl.dialects.builtin.Float64Type | xdsl.dialects.builtin.IndexType but got SSAValue with type i32",
    ):
        SSAValue.get(ssa_value, type=Float64Type | IndexType)

    tensor_type = TensorType(i32, [2, 2])
    ssa_value_tensor = create_ssa_value(tensor_type)

    assert (
        SSAValue.get(ssa_value_tensor, type=TensorType[IntegerType]) == ssa_value_tensor
    )
    assert (
        SSAValue.get(ssa_value_tensor, type=TensorType[IntegerType]).type == tensor_type
    )

    assert (
        SSAValue.get(ssa_value_tensor, type=TensorType[IntegerType | IndexType])
        == ssa_value_tensor
    )
    assert (
        SSAValue.get(ssa_value_tensor, type=TensorType[IntegerType | IndexType]).type
        == tensor_type
    )

    with pytest.raises(
        ValueError,
        match=r"SSAValue.get: Expected xdsl.dialects.builtin.TensorType\[xdsl.dialects.builtin.Float64Type \| xdsl.dialects.builtin.IndexType\] but got SSAValue with type tensor<2x2xi32>.",
    ):
        SSAValue.get(ssa_value_tensor, type=TensorType[Float64Type | IndexType])


def test_ssa_get_on_op():
    op1 = test.TestOp(result_types=[i32])

    assert SSAValue.get(op1).owner == op1
    assert SSAValue.get(op1).type == i32
    assert SSAValue.get(op1, type=IntegerType).owner == op1
    assert SSAValue.get(op1, type=IntegerType).type == i32

    with pytest.raises(
        ValueError,
        match="SSAValue.get: Expected <class 'xdsl.dialects.builtin.IndexType'> but got SSAValue with type i32",
    ):
        SSAValue.get(op1, type=IndexType)

    op2 = test.TestOp(result_types=[i32, i32])
    with pytest.raises(
        ValueError, match="SSAValue.get: expected operation with a single result."
    ):
        SSAValue.get(op2)



from xdsl.dialects import accfg, builtin, test
from xdsl.dialects.builtin import StringAttr


def test_acc_setup():
    one, two = test.TestOp(result_types=[builtin.i32, builtin.i32]).results

    setup1 = accfg.SetupOp([one, two], ["A", "B"], "acc1")
    setup1.verify()

    assert setup1.accelerator == StringAttr("acc1")

    setup2 = accfg.SetupOp(setup1.values, ["A", "B"], setup1.accelerator, setup1)
    setup2.verify()

    assert setup2.accelerator == setup1.accelerator
    assert isinstance(setup2.out_state.type, accfg.StateType)
    assert setup2.out_state.type.accelerator == setup1.accelerator



import abc

from xdsl.backend.register_type import RegisterType
from xdsl.ir import SSAValue


class AssemblyInstructionArg(abc.ABC):
    """
    Abstract base class for arguments to one line of assembly.
    """

    @abc.abstractmethod
    def assembly_str(self) -> str:
        raise NotImplementedError()


class reg(AssemblyInstructionArg):
    """
    A wrapper around SSAValue to be printed in assembly.
    Only valid if the type of the value is a RegisterType.
    """

    value: SSAValue

    def __init__(self, value: SSAValue) -> None:
        self.value = value

    def assembly_str(self) -> str:
        assert isinstance(self.value.type, RegisterType)
        return self.value.type.register_name.data


class square_brackets_reg(AssemblyInstructionArg):
    """
    A wrapper around SSAValue to be printed in assembly.
    Only valid if the type of the value is a RegisterType.
    This class handles the case where a register contains a pointer reference,
    and therefore should be printed within square brackets.
    """

    value: SSAValue

    def __init__(self, value: SSAValue) -> None:
        self.value = value

    def assembly_str(self) -> str:
        assert isinstance(self.value.type, RegisterType)
        return f"[{self.value.type.register_name.data}]"



from collections.abc import Sequence

from xdsl.dialects.builtin import ModuleOp
from xdsl.interactive.passes import AvailablePass
from xdsl.pattern_rewriter import PatternRewriter
from xdsl.traits import HasCanonicalizationPatternsTrait
from xdsl.transforms import individual_rewrite
from xdsl.utils.op_selector import OpSelector


def get_all_possible_rewrites(module: ModuleOp) -> Sequence[AvailablePass]:
    """
    Function that takes a sequence of IndividualRewrite Patterns and a ModuleOp, and
    returns the possible rewrites.
    Issue filed: https://github.com/xdslproject/xdsl/issues/2162
    """

    res: list[AvailablePass] = []

    for op_idx, matched_op in enumerate(module.walk()):
        if (trait := matched_op.get_trait(HasCanonicalizationPatternsTrait)) is None:
            continue

        pattern_by_name = {
            type(pattern).__name__: pattern
            for pattern in trait.get_canonicalization_patterns()
        }

        selector = OpSelector(op_idx, matched_op.name)

        for pattern_name, pattern in pattern_by_name.items():
            cloned_op = selector.get_op(module.clone())
            rewriter = PatternRewriter(cloned_op)
            pattern.match_and_rewrite(cloned_op, rewriter)
            if rewriter.has_done_action:
                res.append(
                    AvailablePass(
                        individual_rewrite.ApplyIndividualRewritePass(
                            op_idx, cloned_op.name, pattern_name
                        ),
                    )
                )

    return res



from __future__ import annotations

import itertools
from collections.abc import Callable, Sequence
from dataclasses import dataclass
from inspect import getfullargspec
from typing import cast

from typing_extensions import TypeVar

from xdsl.ir.affine import AffineConstantExpr, AffineDimExpr, AffineExpr

AffineExprBuilderT = AffineExpr | int

AffineMapBuilderT = (
    Callable[[], tuple[AffineExprBuilderT, ...]]
    | Callable[[AffineExpr], tuple[AffineExprBuilderT, ...]]
    | Callable[[AffineExpr, AffineExpr], tuple[AffineExprBuilderT, ...]]
    | Callable[[AffineExpr, AffineExpr, AffineExpr], tuple[AffineExprBuilderT, ...]]
    | Callable[
        [AffineExpr, AffineExpr, AffineExpr, AffineExpr], tuple[AffineExprBuilderT, ...]
    ]
    | Callable[
        [AffineExpr, AffineExpr, AffineExpr, AffineExpr, AffineExpr],
        tuple[AffineExprBuilderT, ...],
    ]
)

_T = TypeVar("_T")


@dataclass(frozen=True)
class AffineMap:
    """
    AffineMap represents a map from a set of dimensions and symbols to a
    multi-dimensional affine expression.
    """

    num_dims: int
    num_symbols: int
    results: tuple[AffineExpr, ...]

    @staticmethod
    def constant_map(value: int) -> AffineMap:
        return AffineMap(0, 0, (AffineExpr.constant(value),))

    @staticmethod
    def point_map(*values: int) -> AffineMap:
        return AffineMap(0, 0, tuple(AffineExpr.constant(value) for value in values))

    @staticmethod
    def identity(rank: int, symbolic_rank: int = 0) -> AffineMap:
        return AffineMap(
            rank,
            symbolic_rank,
            tuple(AffineExpr.dimension(dim) for dim in range(rank))
            + tuple(AffineExpr.symbol(dim) for dim in range(symbolic_rank)),
        )

    @staticmethod
    def minor_identity(num_dims: int, num_results: int) -> AffineMap:
        """
        Returns an identity affine map (d0, ..., dn) -> (dp, ..., dn) on the most minor
        dimensions.

        Corresponds to MLIR's `AffineMap::getMinorIdentityMap`.
        """
        if num_dims < num_results:
            raise ValueError(
                f"Dimension mismatch, expected dims {num_dims} to be greater than or "
                f"equal to results {num_results}."
            )

        return AffineMap(
            num_dims,
            0,
            tuple(AffineDimExpr(d) for d in range(num_dims - num_results, num_dims)),
        )

    @staticmethod
    def transpose_map() -> AffineMap:
        """
        Returns the map transposing a 2D matrix: `(i, j) -> (j, i)`.
        """
        return AffineMap(2, 0, (AffineExpr.dimension(1), AffineExpr.dimension(0)))

    @staticmethod
    def empty() -> AffineMap:
        return AffineMap(0, 0, ())

    @staticmethod
    def from_callable(
        func: AffineMapBuilderT, *, dim_symbol_split: tuple[int, int] | None = None
    ) -> AffineMap:
        """
        Creates an `AffineMap` by calling the function provided. If `dim_symbol_split` is
        not provided or `None`, then all parameters are treated as dimension expressions.
        If `dim_symbol_split` is provided, `func` is expected to have the same number of
        arguments as the sum of elements of `dim_symbol_split`.

        3D Identity:
        ```
        AffineMap.from_callable(lambda i, j, k: (i, j, k))
        ```
        Constant:
        ```
        AffineMap.from_callable(lambda i, j: (0, 0))
        ```
        Mix of dimensions and symbols:
        ```
        AffineMap.from_callable(lambda i, p: (p, i), dim_symbol_split=(1,1))
        ```
        """
        sig = getfullargspec(func)
        num_args = len(sig.args)
        if dim_symbol_split is None:
            num_dims = num_args
            num_symbols = 0
        else:
            num_dims, num_symbols = dim_symbol_split
            if num_args != num_dims + num_symbols:
                raise ValueError(
                    f"Argument count mismatch in AffineMap.from_callable: {num_args} != "
                    f"{num_dims} + {num_symbols}"
                )
        dim_exprs = [AffineExpr.dimension(dim) for dim in range(num_dims)]
        sym_exprs = [AffineExpr.symbol(sym) for sym in range(num_symbols)]
        result_exprs = func(*dim_exprs, *sym_exprs)
        results_tuple = tuple(
            AffineExpr.constant(r) if isinstance(r, int) else r for r in result_exprs
        )
        return AffineMap(num_dims, num_symbols, results_tuple)

    def replace_dims_and_symbols(
        self,
        new_dims: Sequence[AffineExpr],
        new_symbols: Sequence[AffineExpr],
        result_num_dims: int,
        result_num_symbols: int,
    ) -> AffineMap:
        """
        This method substitutes any uses of dimensions and symbols (e.g. dim#0 with
        dimReplacements[0]) in subexpressions and returns the modified expression
        mapping.  Because this can be used to eliminate dims and symbols, the client
        needs to specify the number of dims and symbols in the result.

        The returned map always has the same number of results.
        """

        return AffineMap(
            result_num_dims,
            result_num_symbols,
            tuple(
                expr.replace_dims_and_symbols(new_dims, new_symbols)
                for expr in self.results
            ),
        )

    def compose(self, other: AffineMap) -> AffineMap:
        """
        Returns the `AffineMap` resulting from composing `self` with `other`.

        The resulting `AffineMap` has as many dimensions as `other` and as many symbols
        as the concatenation of `self` and `other` (in which case the symbols of `self`
        come first).

        Prerequisites: The maps are composable, i.e. that the number of dimensions of
        `self` matches the number of results of `other`.

        Example:
        ```
        map1: (d0, d1)[s0, s1] -> (d0 + 1 + s1, d1 - 1 - s0)
        map2: (d0)[s0] -> (d0 + s0, d0 - s0)
        map1.compose(map2): (d0)[s0, s1, s2] -> (d0 + s1 + s2 + 1, d0 - s0 - s2 - 1)
        ```
        """
        if self.num_dims != len(other.results):
            raise ValueError(
                "Cannot compose AffineMaps with mismatching dimensions and results: "
                "self.num_dims != len(map.results) "
                f"({self.num_dims} != {len(other.results)})"
            )

        num_dims = other.num_dims
        num_symbols = self.num_symbols + other.num_symbols

        new_dims = tuple(AffineExpr.dimension(d) for d in range(num_dims))
        new_symbols = tuple(
            AffineExpr.symbol(s) for s in range(self.num_symbols, num_symbols)
        )

        new_map = other.replace_dims_and_symbols(
            new_dims, new_symbols, num_dims, num_symbols
        )

        results = tuple(expr.compose(new_map) for expr in self.results)
        return AffineMap(
            num_dims=num_dims,
            num_symbols=num_symbols,
            results=results,
        )

    def inverse_permutation(self) -> AffineMap | None:
        """
        Returns a map of codomain to domain dimensions such that the first
        codomain dimension for a particular domain dimension is selected.
        Returns an empty map if the input map is empty. Returns null map (not
        empty map) if the map is not invertible (i.e. the map does not contain
        a subset that is a permutation of full domain rank).

        Prerequisites: The map should have no symbols.

        Example:
           (d0, d1, d2) -> (d1, d1, d0, d2, d1, d2, d1, d0)
                             0       2   3
        returns:
           (d0, d1, d2, d3, d4, d5, d6, d7) -> (d2, d0, d3)
        """
        if self.num_symbols != 0:
            raise ValueError(
                f"Cannot invert AffineMap with symbols: {self.num_symbols}"
            )
        found_dims = [-1] * self.num_dims

        for i, expr in enumerate(self.results):
            match expr:
                case AffineDimExpr():
                    if found_dims[expr.position] == -1:
                        found_dims[expr.position] = i
                case _:
                    continue

        if -1 in found_dims:
            return None

        results = tuple(AffineExpr.dimension(i) for i in found_dims)
        return AffineMap(
            num_dims=len(self.results),
            num_symbols=0,
            results=results,
        )

    def inverse_and_broadcast_projected_permutation(self):
        """
        If `self` is a projected permutation, with possible constant 0 expression
        results, returns the inverse permutation.

        Examples:
        ```
        (d0, d1, d2) -> (d2, d1, d0) => (d0, d1, d2) -> (d2, d1, d0)
        (d0, d1, d2) -> (d1, d0)     => (d0, d1)     -> (d1, d0, 0)
        (d0, d1, d2) -> (d1, 0, d0)  => (d0, d1, d2) -> (d2, d0, 0)
        ```

        Equivalent to `inverseAndBroadcastProjectedPermutation` in MLIR.
        """
        assert self.is_projected_permutation(allow_zero_in_results=True), f"{self}"
        results = cast(tuple[AffineConstantExpr | AffineDimExpr, ...], self.results)
        zero = AffineExpr.constant(0)
        # Start with all the results as 0.
        exprs = [zero] * self.num_dims
        for i, res in enumerate(results):
            if isinstance(res, AffineDimExpr):
                # Reverse each dimension existing in the original map result.
                exprs[res.position] = AffineExpr.dimension(i)

        return AffineMap(len(self.results), 0, tuple(exprs))

    def eval(self, dims: Sequence[int], symbols: Sequence[int]) -> tuple[int, ...]:
        """Evaluate the AffineMap given the values of dimensions and symbols."""
        assert len(dims) == self.num_dims, f"{len(dims)}, {self.num_dims}"
        assert len(symbols) == self.num_symbols, f"{len(symbols)}, {self.num_symbols}"
        return tuple(expr.eval(dims, symbols) for expr in self.results)

    def drop_dims(self, unused_dims: Sequence[bool]) -> AffineMap:
        """
        Given a sequence of `unused_dims` indicating the input dimensions to drop,
        return a new map only with the new dimensions. The results of `self` must be a
        subset of the dimensions in `selectors`. The remaining dimensions are remapped
        to the remaining number.

        Examples:
        ```
        (d0, d1, d2) -> (d1, d2) with [T,F,F] gives (d0, d1) -> (d0, d1)
        (d0, d1, d2) -> (d2, d2) with [F,T,F] gives (d0, d1) -> (d1, d1)
        ```

        Corresponds to MLIR's `compressDims`.
        """
        if len(unused_dims) != self.num_dims:
            raise ValueError(
                f"Invalid `unused_dims`, expected {self.num_dims} `bool` values, got "
                f"{len(unused_dims)}"
            )

        result_num_dims = sum(not dim for dim in unused_dims)
        new_dims = tuple(
            AffineExpr.dimension(dim)
            for dim in itertools.accumulate((not dim for dim in unused_dims), initial=0)
        )
        new_symbols = tuple(AffineExpr.symbol(s) for s in range(self.num_symbols))

        return self.replace_dims_and_symbols(
            new_dims, new_symbols, result_num_dims, self.num_symbols
        )

    def drop_results(self, unused_results: Sequence[bool]) -> AffineMap:
        """
        Given a sequence of `unused_results` indicating the results to drop,
        return a new map only with the new results.

        Examples:
        ```
        (d0, d1, d2) -> (d1, d2) with [T,F] gives (d0, d1, d2) -> (d1)
        (d0, d1, d2) -> (d1, d2) with [F,T] gives (d0, d1, d2) -> (d1)
        ```

        Corresponds to MLIR's `dropResults`, but passing a mask instead of integer
        indices to drop.
        """
        if len(unused_results) != len(self.results):
            raise ValueError(
                f"Invalid `unused_results`, expected {len(self.results)} `bool` values, got "
                f"{len(unused_results)}"
            )

        return AffineMap(
            self.num_dims,
            self.num_symbols,
            tuple(
                result
                for (mask, result) in zip(unused_results, self.results)
                if not mask
            ),
        )

    def used_dims(self) -> set[int]:
        """
        Return all dimensions used in the map as a set

        Example:
        ```
        (d0, d1) -> (d0) gives {d0}
        (d0, d1, d2) -> (d0, d2) gives {d0, d2}
        ```
        """
        return {
            expr.position
            for res_expr in self.results
            for expr in res_expr.dfs()
            if isinstance(expr, AffineDimExpr)
        }

    def unused_dims(self) -> set[int]:
        """
        Return all dimensions not used in the map as a set

        Example:
        ```
        (d0, d1) -> (d0) gives {d1}
        (d0, d1, d2, d3) -> (d0, d2) gives {d1, d3}
        ```
        """
        return self.used_dims().symmetric_difference(range(self.num_dims))

    def used_dims_bit_vector(self) -> tuple[bool, ...]:
        """
        Return a tuple of bools with the i-th entry being True if the i-th dimension is
        used in the map, otherwise it is False.

        Example:
        ```
        (d0, d1) -> (d0) gives (True, False)
        (d0, d1, d2) -> (d0, d2) gives (True, False, True)
        ```
        """
        used_dims = self.used_dims()
        return tuple(dim in used_dims for dim in range(self.num_dims))

    def unused_dims_bit_vector(self) -> tuple[bool, ...]:
        """
        Return a tuple of bools with the i-th entry being True if the i-th dimension is
        not used in the map, otherwise it is False.

        Example:
        ```
        (d0, d1) -> (d0) gives (True, False)
        (d0, d1, d2) -> (d0, d2) gives (True, False, True)
        ```
        """
        used_dims = self.used_dims()
        return tuple(dim not in used_dims for dim in range(self.num_dims))

    def is_minor_identity(self) -> bool:
        """
        Returns True if
        1. there are at most `self.num_dims` results,
        2. `self.num_symbols` is zero, and
        3. `self.results` are the last dimensions, in order.

        For example, `(d0, d1, d2) -> (d1, d2)` is a minor identity map.

        Corresponds to MLIR's `AffineMap::isMinorIdentity`.
        """
        num_results = len(self.results)
        return (
            not self.num_symbols
            and num_results <= self.num_dims
            and all(
                isinstance(r, AffineDimExpr) and d == r.position
                for d, r in zip(
                    range(self.num_dims - num_results, self.num_dims),
                    self.results,
                    strict=True,
                )
            )
        )

    def is_projected_permutation(self, allow_zero_in_results: bool = False) -> bool:
        """
        Returns True if the AffineMap represents a subset (i.e. a projection) of a
        symbol-less permutation map. `allow_zero_in_results` allows projected
        permutation maps with constant zero result expressions.

        Examples:
        ```
        no_zeros = (d0, d1, d2) -> (d1, d0)
        with_zeros = (d0, d1, d2) -> (d1, 0, d0)
        ```

        Equivalent to `isProjectedPermutation` in MLIR.
        """
        if self.num_symbols:
            return False

        # Having more results than inputs means that results have duplicated dims or
        # zeros that can't be mapped to input dims.
        if len(self.results) > self.num_dims:
            return False

        seen = [False] * self.num_dims
        # A projected permutation can have, at most, only one instance of each input
        # dimension in the result expressions. Zeros are allowed as long as the
        # number of result expressions is lower or equal than the number of input
        # expressions.
        for expr in self.results:
            if isinstance(expr, AffineDimExpr):
                if seen[expr.position]:
                    return False
                seen[expr.position] = True
            else:
                if (
                    not allow_zero_in_results
                    or not isinstance(expr, AffineConstantExpr)
                    or expr.value != 0
                ):
                    return False

        # Results are either dims or zeros and zeros can be mapped to input dims.
        return True

    def apply_permutation(self, source: Sequence[_T]) -> tuple[_T, ...]:
        """
        Assert that `self` represents a projected permutation, and apply the permutation
        to `source`.
        The number of inputs must match the size of the source.

        Example:
        ```
        map = (d0, d1, d2) -> (d1, d0)
        source = [10, 20, 30]
        result = [20, 10]
        ```

        Equivalent to `applyPermutationMap` in MLIR.
        """
        assert self.is_projected_permutation(), "Map must be a projected permutation"
        assert self.num_dims == len(source), "Number of inputs must match source size"
        results = cast(Sequence[AffineDimExpr], self.results)
        return tuple(source[expr.position] for expr in results)

    def __str__(self) -> str:
        # Create comma seperated list of dims.
        dims = ["d" + str(i) for i in range(self.num_dims)]
        dims = ", ".join(dims)
        # Create comma seperated list of symbols.
        syms = ["s" + str(i) for i in range(self.num_symbols)]
        syms = ", ".join(syms)
        # Create comma seperated list of results.
        results = ", ".join(str(expr) for expr in self.results)
        if self.num_symbols == 0:
            return f"({dims}) -> ({results})"
        return f"({dims})[{syms}] -> ({results})"



from dataclasses import dataclass

from xdsl.backend.riscv.register_allocation import RegisterAllocatorLivenessBlockNaive
from xdsl.backend.riscv.register_stack import RiscvRegisterStack
from xdsl.context import Context
from xdsl.dialects import riscv_func
from xdsl.dialects.builtin import ModuleOp
from xdsl.passes import ModulePass


@dataclass(frozen=True)
class RISCVRegisterAllocation(ModulePass):
    """
    Allocates unallocated registers in the module.
    """

    name = "riscv-allocate-registers"

    allocation_strategy: str = "LivenessBlockNaive"

    add_regalloc_stats: bool = False
    """
    Inserts a comment with register allocation info in the IR.
    """

    allow_infinite: bool = False
    """
    Whether to allow using infinite registers during register allocation.
    """

    def apply(self, ctx: Context, op: ModuleOp) -> None:
        allocator_strategies = {
            "LivenessBlockNaive": RegisterAllocatorLivenessBlockNaive,
        }

        if self.allocation_strategy not in allocator_strategies:
            raise ValueError(
                f"Unknown register allocation strategy {self.allocation_strategy}. "
                f"Available allocation types: {allocator_strategies.keys()}"
            )

        for inner_op in op.walk():
            if isinstance(inner_op, riscv_func.FuncOp):
                register_stack = RiscvRegisterStack.get(
                    allow_infinite=self.allow_infinite
                )
                allocator = allocator_strategies[self.allocation_strategy](
                    register_stack
                )
                allocator.allocate_func(
                    inner_op, add_regalloc_stats=self.add_regalloc_stats
                )



from collections.abc import Sequence

import pytest
from typing_extensions import TypeVar

from xdsl.dialects.builtin import StringAttr
from xdsl.ir import Attribute
from xdsl.irdl import (
    AnyAttr,
    AnyInt,
    AttrConstraint,
    BaseAttr,
    ConstraintContext,
    IntVarConstraint,
    RangeConstraint,
    RangeLengthConstraint,
    RangeOf,
    VarConstraint,
)
from xdsl.utils.exceptions import VerifyException


class AnyRangeConstraint(RangeConstraint):
    """Constraint for testing default infer"""

    def verify(
        self, attrs: Sequence[Attribute], constraint_context: ConstraintContext
    ) -> None:
        return

    def verify_length(self, length: int, constraint_context: ConstraintContext) -> None:
        return

    def mapping_type_vars(
        self, type_var_mapping: dict[TypeVar, AttrConstraint]
    ) -> RangeConstraint:
        return self


def test_failing_inference():
    with pytest.raises(
        ValueError, match="Cannot infer range from constraint AnyRangeConstraint()"
    ):
        AnyRangeConstraint().infer(ConstraintContext(), length=None)


def test_range_of_variables():
    attr_constr = VarConstraint("ATTR", AnyAttr())
    range_constr = RangeOf(attr_constr)
    assert range_constr.variables() == {"ATTR"}
    len_constr = IntVarConstraint("LENGTH", AnyInt())
    assert range_constr.of_length(len_constr).variables() == {"ATTR", "LENGTH"}


def test_verify_range_length_constraint():
    hello = StringAttr("hello")
    world = StringAttr("world")
    attr_constr = VarConstraint("ATTR", BaseAttr(StringAttr))
    len_constr = IntVarConstraint("LENGTH", AnyInt())
    range_len_constr = RangeLengthConstraint(RangeOf(attr_constr), len_constr)
    with pytest.raises(
        VerifyException,
        match='''attribute "hello" expected from variable 'ATTR', but got "world"''',
    ):
        range_len_constr.verify((hello, world), ConstraintContext())
    range_len_constr.verify((world, world), ConstraintContext())

    with pytest.raises(
        VerifyException,
        match="incorrect length for range variable",
    ):
        range_len_constr.verify(
            (world, world, world), ConstraintContext(_int_variables={"LENGTH": 2})
        )

    # verify_length
    range_len_constr.verify_length(2, ConstraintContext())

    with pytest.raises(
        VerifyException,
        match="incorrect length for range variable",
    ):
        range_len_constr.verify_length(
            3, ConstraintContext(_int_variables={"LENGTH": 2})
        )

    # variables
    assert range_len_constr.variables() == {"ATTR", "LENGTH"}

    # variables_from_length
    assert range_len_constr.variables_from_length() == {"LENGTH"}

    # can_infer
    assert not range_len_constr.can_infer(set(), length_known=True)
    assert not range_len_constr.can_infer(set(), length_known=False)
    assert not range_len_constr.can_infer({"ATTR"}, length_known=False)
    assert range_len_constr.can_infer({"ATTR"}, length_known=True)
    assert range_len_constr.can_infer({"ATTR", "LENGTH"}, length_known=False)

    # infer
    assert range_len_constr.infer(
        ConstraintContext(_variables={"ATTR": world}), length=2
    ) == (
        world,
        world,
    )
    assert range_len_constr.infer(
        ConstraintContext(_variables={"ATTR": world}, _int_variables={"LENGTH": 3}),
        length=None,
    ) == (world, world, world)



from abc import ABC, abstractmethod

from xdsl.backend.assembly_printer import AssemblyPrinter, OneLineAssemblyPrintable
from xdsl.dialects.builtin import StringAttr
from xdsl.ir import Operation, SSAValue
from xdsl.irdl import (
    IRDLOperation,
    irdl_op_definition,
    operand_def,
    opt_attr_def,
    prop_def,
    result_def,
)

from .assembly import AssemblyInstructionArg, reg
from .register import IntRegisterType


class ARMOperation(IRDLOperation, OneLineAssemblyPrintable, ABC):
    """
    Base class for operations that can be a part of ARM assembly printing.
    """


class ARMInstruction(ARMOperation, ABC):
    """
    Base class for operations that can be a part of x86 assembly printing. Must
    represent an instruction in the x86 instruction set.
    The name of the operation will be used as the x86 assembly instruction name.
    """

    comment = opt_attr_def(StringAttr)
    """
    An optional comment that will be printed along with the instruction.
    """

    @abstractmethod
    def assembly_line_args(self) -> tuple[AssemblyInstructionArg | None, ...]:
        """
        The arguments to the instruction, in the order they should be printed in the
        assembly.
        """
        raise NotImplementedError()

    def assembly_instruction_name(self) -> str:
        """
        By default, the name of the instruction is the same as the name of the operation.
        """

        return self.name.split(".")[-1]

    def assembly_line(self) -> str | None:
        # default assembly code generator
        instruction_name = self.assembly_instruction_name()
        arg_str = ", ".join(
            arg.assembly_str() for arg in self.assembly_line_args() if arg is not None
        )
        return AssemblyPrinter.assembly_line(instruction_name, arg_str, self.comment)


@irdl_op_definition
class DSMovOp(ARMInstruction):
    """
    Copies the value of s into d.

    See external [documentation](https://developer.arm.com/documentation/dui0473/m/arm-and-thumb-instructions/mov).
    """

    name = "arm.ds.mov"

    d = result_def(IntRegisterType)
    s = operand_def(IntRegisterType)
    assembly_format = "$s attr-dict `:` `(` type($s) `)` `->` type($d)"

    def __init__(
        self,
        d: IntRegisterType,
        s: Operation | SSAValue,
        *,
        comment: str | StringAttr | None = None,
    ):
        if isinstance(comment, str):
            comment = StringAttr(comment)

        super().__init__(
            operands=(s,),
            attributes={
                "comment": comment,
            },
            result_types=(d,),
        )

    def assembly_line_args(self):
        return (reg(self.d), reg(self.s))


@irdl_op_definition
class GetRegisterOp(ARMOperation):
    """
    This instruction allows us to create an SSAValue for a given register name.
    """

    name = "arm.get_register"

    result = result_def(IntRegisterType)
    assembly_format = "attr-dict `:` type($result)"

    def __init__(self, register_type: IntRegisterType):
        super().__init__(result_types=[register_type])

    def assembly_line(self):
        return None


@irdl_op_definition
class DSSMulOp(ARMInstruction):
    """
    Multiplies the values in s1 and s2 and stores the result in d.

    See external [documentation](https://developer.arm.com/documentation/ddi0597/2024-06/Base-Instructions/MUL--MULS--Multiply-?lang=en).
    """

    name = "arm.dss.mul"

    d = result_def(IntRegisterType)
    s1 = operand_def(IntRegisterType)
    s2 = operand_def(IntRegisterType)
    assembly_format = (
        "$s1 `,` $s2 attr-dict `:` `(` type($s1) `,` type($s2) `)` `->` type($d)"
    )

    def __init__(
        self,
        s1: Operation | SSAValue,
        s2: Operation | SSAValue,
        *,
        d: IntRegisterType,
        comment: str | StringAttr | None = None,
    ):
        if isinstance(comment, str):
            comment = StringAttr(comment)

        super().__init__(
            operands=(s1, s2),
            attributes={
                "comment": comment,
            },
            result_types=(d,),
        )

    def assembly_line_args(self):
        return (reg(self.d), reg(self.s1), reg(self.s2))


@irdl_op_definition
class LabelOp(ARMOperation):
    """
    The label operation is used to emit text labels (e.g. loop:) that are used
    as branch, unconditional jump targets and symbol offsets.

    See external [documentation](https://developer.arm.com/documentation/dui0801/l/Symbols--Literals--Expressions--and-Operators/Labels).
    """

    name = "arm.label"
    label = prop_def(StringAttr)
    comment = opt_attr_def(StringAttr)

    assembly_format = "$label attr-dict"

    def __init__(
        self,
        label: str | StringAttr,
        *,
        comment: str | StringAttr | None = None,
    ):
        if isinstance(label, str):
            label = StringAttr(label)
        if isinstance(comment, str):
            comment = StringAttr(comment)

        super().__init__(
            attributes={
                "label": label,
                "comment": comment,
            },
        )

    def assembly_line(self) -> str | None:
        return AssemblyPrinter.append_comment(f"{self.label.data}:", self.comment)


@irdl_op_definition
class CmpRegOp(ARMInstruction):
    """
    Compare (register) subtracts an optionally-shifted register value from a register value.
    It updates the condition flags based on the result, and discards the result.

    See external [documentation](https://developer.arm.com/documentation/ddi0597/2024-12/Base-Instructions/CMP--register---Compare--register--?lang=en).
    """

    name = "arm.cmp"
    s1 = operand_def(IntRegisterType)
    s2 = operand_def(IntRegisterType)

    assembly_format = "$s1 `,` $s2 attr-dict `:` `(` type($s1) `,` type($s2) `)`"

    def __init__(
        self,
        s1: Operation | SSAValue,
        s2: Operation | SSAValue,
        *,
        comment: str | StringAttr | None = None,
    ):
        if isinstance(comment, str):
            comment = StringAttr(comment)

        super().__init__(
            operands=(s1, s2),
            attributes={
                "comment": comment,
            },
        )

    def assembly_line_args(self):
        return (reg(self.s1), reg(self.s2))



from __future__ import annotations

import abc
from collections.abc import Iterator
from dataclasses import dataclass, field
from enum import Enum, auto
from typing import TYPE_CHECKING, final

from typing_extensions import TypeVar

from xdsl.utils.exceptions import VerifyException

if TYPE_CHECKING:
    from xdsl.dialects.builtin import StringAttr, SymbolRefAttr
    from xdsl.ir import Attribute, Operation, Region, SSAValue
    from xdsl.pattern_rewriter import RewritePattern


@dataclass(frozen=True)
class OpTrait:
    """
    A trait attached to an operation definition.
    Traits can be used to define operation invariants, additional semantic information,
    or to group operations that have similar properties.
    Note that traits are the merge of traits and interfaces in MLIR.
    """

    def verify(self, op: Operation) -> None:
        """Check that the operation satisfies the trait requirements."""
        pass


OpTraitInvT = TypeVar("OpTraitInvT", bound=OpTrait)


class ConstantLike(OpTrait):
    """
    Operation known to be constant-like.

    See external [documentation](https://mlir.llvm.org/doxygen/classmlir_1_1OpTrait_1_1ConstantLike.html).
    """


@dataclass(frozen=True)
class HasParent(OpTrait):
    """Constraint the operation to have a specific parent operation."""

    op_types: tuple[type[Operation], ...]

    def __init__(self, head_param: type[Operation], *tail_params: type[Operation]):
        object.__setattr__(self, "op_types", (head_param, *tail_params))

    def verify(self, op: Operation) -> None:
        parent = op.parent_op()
        # Don't check parent when op is detached
        if parent is None:
            return
        if isinstance(parent, self.op_types):
            return
        if len(self.op_types) == 1:
            raise VerifyException(
                f"'{op.name}' expects parent op '{self.op_types[0].name}'"
            )
        names = ", ".join(f"'{p.name}'" for p in self.op_types)
        raise VerifyException(f"'{op.name}' expects parent op to be one of {names}")


@dataclass(frozen=True)
class HasAncestor(OpTrait):
    """
    Constraint the operation to have a specific operation as ancestor, i.e. transitive
    parent.
    """

    op_types: tuple[type[Operation], ...]

    def __init__(self, head_param: type[Operation], *tail_params: type[Operation]):
        object.__setattr__(self, "op_types", (head_param, *tail_params))

    def verify(self, op: Operation) -> None:
        if self.get_ancestor(op) is None:
            if len(self.op_types) == 1:
                raise VerifyException(
                    f"'{op.name}' expects ancestor op '{self.op_types[0].name}'"
                )
            names = ", ".join(f"'{p.name}'" for p in self.op_types)
            raise VerifyException(
                f"'{op.name}' expects ancestor op to be one of {names}"
            )

    def walk_ancestors(self, op: Operation) -> Iterator[Operation]:
        """Iterates over the ancestors of an operation, including the input"""
        curr = op
        yield curr
        while (curr := curr.parent_op()) is not None:
            yield curr

    def get_ancestor(self, op: Operation) -> Operation | None:
        ancestors = self.walk_ancestors(op)
        matching_ancestors = (a for a in ancestors if isinstance(a, self.op_types))
        return next(matching_ancestors, None)


class IsTerminator(OpTrait):
    """
    This trait provides verification and functionality for operations that are
    known to be terminators.

    See external [documentation](https://mlir.llvm.org/docs/Traits/#terminator).
    """

    def verify(self, op: Operation) -> None:
        """Check that the operation satisfies the IsTerminator trait requirements."""
        if op.parent is not None and op.parent.last_op != op:
            raise VerifyException(
                f"'{op.name}' must be the last operation in its parent block"
            )


class NoTerminator(OpTrait):
    """
    Allow an operation to have single block regions with no terminator.

    See external [documentation](https://mlir.llvm.org/docs/Traits/#terminator).
    """

    def verify(self, op: Operation) -> None:
        for region in op.regions:
            if len(region.blocks) > 1:
                raise VerifyException(
                    f"'{op.name}' does not contain single-block regions"
                )


@dataclass(frozen=True)
class SingleBlockImplicitTerminator(OpTrait):
    """
    Checks the existence of the specified terminator to an operation which has
    single-block regions.
    The conditions for the implicit creation of the terminator depend on the operation
    and occur during its creation using the `ensure_terminator` method.

    This should be fully compatible with MLIR's Trait.

    See external [documentation](https://mlir.llvm.org/docs/Traits/#single-block-with-implicit-terminator).
    """

    op_type: type[Operation]

    def verify(self, op: Operation) -> None:
        for region in op.regions:
            if len(region.blocks) > 1:
                raise VerifyException(
                    f"'{op.name}' does not contain single-block regions"
                )
            for block in region.blocks:
                if (last_op := block.last_op) is None:
                    raise VerifyException(
                        f"'{op.name}' contains empty block instead of at least "
                        f"terminating with {self.op_type.name}"
                    )

                if not isinstance(last_op, self.op_type):
                    raise VerifyException(
                        f"'{op.name}' terminates with operation {last_op.name} "
                        f"instead of {self.op_type.name}"
                    )


def ensure_terminator(op: Operation, trait: SingleBlockImplicitTerminator) -> None:
    """
    Method that helps with the creation of an implicit terminator.
    This should be explicitly called during the creation of an operation that has the
    SingleBlockImplicitTerminator trait.
    """

    for region in op.regions:
        if len(region.blocks) > 1:
            raise VerifyException(f"'{op.name}' does not contain single-block regions")

        from xdsl.dialects.builtin import UnregisteredOp

        for block in region.blocks:
            if (
                (last_op := block.last_op) is not None
                and not isinstance(last_op, UnregisteredOp)
                and last_op.has_trait(IsTerminator)
                and not isinstance(last_op, trait.op_type)
            ):
                raise VerifyException(
                    f"'{op.name}' terminates with operation {last_op.name} "
                    f"instead of {trait.op_type.name}"
                )

    from xdsl.builder import ImplicitBuilder
    from xdsl.ir import Block

    for region in op.regions:
        if not region.blocks:
            region.add_block(Block())

        for block in region.blocks:
            if (last_op := block.last_op) is None or not last_op.has_trait(
                IsTerminator, value_if_unregistered=False
            ):
                with ImplicitBuilder(block):
                    trait.op_type.create()


class IsolatedFromAbove(OpTrait):
    """
    Constrains the contained operations to use only values defined inside this
    operation.

    This should be fully compatible with MLIR's Trait.

    See external [documentation](https://mlir.llvm.org/docs/Traits/#isolatedfromabove).
    """

    def verify(self, op: Operation) -> None:
        # Start by checking all the passed operation's regions
        regions: list[Region] = list(op.regions)

        # While regions are left to check
        while regions:
            # Pop the first one
            region = regions.pop()
            # Check every block of the region
            for block in region.blocks:
                # Check every operation of the block
                for child_op in block.ops:
                    # Check every operand of the operation
                    for operand in child_op.operands:
                        # The operand must not be defined out of the IsolatedFromAbove op.
                        if not op.is_ancestor(operand.owner):
                            raise VerifyException(
                                "Operation using value defined out of its "
                                f"IsolatedFromAbove parent: {child_op}"
                            )
                    # Check nested regions too; unless the operation is IsolatedFromAbove
                    # too; in which case it will check itself.
                    if not child_op.has_trait(IsolatedFromAbove):
                        regions += child_op.regions


class SymbolUserOpInterface(OpTrait, abc.ABC):
    """
    Used to represent operations that reference Symbol operations. This provides the
    ability to perform safe and efficient verification of symbol uses, as well as
    additional functionality.

    See external [documentation](https://mlir.llvm.org/docs/Interfaces/#symbolinterfaces).
    """

    @abc.abstractmethod
    def verify(self, op: Operation) -> None:
        """
        This method should be adapted to the requirements of specific symbol users per
        operation.

        It corresponds to the verifySymbolUses in upstream MLIR.
        """
        raise NotImplementedError()


class SymbolTable(OpTrait):
    """
    SymbolTable operations are containers for Symbol operations. They offer lookup
    functionality for Symbols, and enforce unique symbols amongst its children.

    A SymbolTable operation is constrained to have a single single-block region.
    """

    def verify(self, op: Operation):
        # import builtin here to avoid circular import
        from xdsl.dialects.builtin import StringAttr

        if len(op.regions) != 1:
            raise VerifyException(
                "Operations with a 'SymbolTable' must have exactly one region"
            )
        if len(op.regions[0].blocks) != 1:
            raise VerifyException(
                "Operations with a 'SymbolTable' must have exactly one block"
            )
        block = op.regions[0].blocks[0]
        met_names: set[StringAttr] = set()
        for o in block.ops:
            if (sym_name := o.get_attr_or_prop("sym_name")) is None:
                continue
            if not isinstance(sym_name, StringAttr):
                continue
            if sym_name in met_names:
                raise VerifyException(f'Redefinition of symbol "{sym_name.data}"')
            met_names.add(sym_name)

    @staticmethod
    def lookup_symbol(
        op: Operation, name: str | StringAttr | SymbolRefAttr
    ) -> Operation | None:
        """
        Lookup a symbol by reference, starting from a specific operation's closest
        SymbolTable parent.
        """
        # import builtin here to avoid circular import
        from xdsl.dialects.builtin import StringAttr, SymbolRefAttr

        anchor: Operation | None = op
        while anchor is not None and not anchor.has_trait(SymbolTable):
            anchor = anchor.parent_op()
        if anchor is None:
            raise ValueError(f"Operation {op} has no SymbolTable ancestor")
        if isinstance(name, str | StringAttr):
            name = SymbolRefAttr(name)
        for o in anchor.regions[0].block.ops:
            if (
                sym_interface := o.get_trait(SymbolOpInterface)
            ) is not None and sym_interface.get_sym_attr_name(o) == name.root_reference:
                if not name.nested_references:
                    return o
                nested_root, *nested_references = name.nested_references.data
                nested_name = SymbolRefAttr(nested_root, nested_references)
                return SymbolTable.lookup_symbol(o, nested_name)
        return None

    @staticmethod
    def insert_or_update(
        symbol_table_op: Operation, symbol_op: Operation
    ) -> Operation | None:
        """
        This takes a symbol_table_op and a symbol_op. It looks if another operation
        inside symbol_table_op already defines symbol_ops symbol. If another operation
        is found, it replaces that operation with symbol_op. Otherwise, symbol_op is
        inserted at the end of symbol_table_op.

        This method returns the operation that was replaced or None if no operation
        was replaced.
        """
        trait = symbol_op.get_trait(SymbolOpInterface)

        if trait is None:
            raise ValueError(
                "Passed symbol_op does not have the SymbolOpInterface trait"
            )

        symbol_name = trait.get_sym_attr_name(symbol_op)

        if symbol_name is None:
            raise ValueError("Passed symbol_op does not have a symbol attribute name")

        tbl_trait = symbol_table_op.get_trait(SymbolTable)

        if tbl_trait is None:
            raise ValueError("Passed symbol_table_op does not have a SymbolTable trait")

        defined_symbol = tbl_trait.lookup_symbol(symbol_table_op, symbol_name)

        if defined_symbol is None:
            symbol_table_op.regions[0].blocks[0].add_op(symbol_op)
            return None
        else:
            parent = defined_symbol.parent
            assert parent is not None
            parent.insert_op_after(symbol_op, defined_symbol)
            parent.detach_op(defined_symbol)
            return defined_symbol


class SymbolOpInterface(OpTrait):
    """
    A `Symbol` is a named operation that resides immediately within a region that defines
    a `SymbolTable` (TODO). A Symbol operation should use the SymbolOpInterface interface to
    provide the necessary verification and accessors.

    A Symbol operation may be optional or not. If - the default - it is not optional,
    a `sym_name` attribute of type StringAttr is required. If it is optional,
    the attribute is optional too.

    xDSL offers OptionalSymbolOpInterface as an always-optional SymbolOpInterface helper.

    More requirements are defined in MLIR; Please see MLIR documentation for Symbol and
    SymbolTable for the requirements that are upcoming in xDSL.

    See external [documentation](https://mlir.llvm.org/docs/SymbolsAndSymbolTables/#symbol).
    """

    def get_sym_attr_name(self, op: Operation) -> StringAttr | None:
        """
        Returns the symbol of the operation, if any
        """
        # import builtin here to avoid circular import
        from xdsl.dialects.builtin import StringAttr

        sym_name = op.get_attr_or_prop("sym_name")
        if sym_name is None and self.is_optional_symbol(op):
            return None
        if not isinstance(sym_name, StringAttr):
            raise VerifyException(
                f'Operation {op.name} must have a "sym_name" attribute of type '
                f"`StringAttr` to conform to {SymbolOpInterface.__name__}"
            )
        return sym_name

    def is_optional_symbol(self, op: Operation) -> bool:
        """
        Returns true if this operation optionally defines a symbol based on the
        presence of the symbol name.
        """
        return False

    def verify(self, op: Operation) -> None:
        # This helper has the same behaviour, so we reuse it as a verifier.That is, it
        # raises a VerifyException iff this operation is a non-optional symbol *and*
        # there is no "sym_name" attribute or property.
        self.get_sym_attr_name(op)


class OptionalSymbolOpInterface(SymbolOpInterface):
    """
    Helper interface specialization for an optional SymbolOpInterface.
    """

    def is_optional_symbol(self, op: Operation) -> bool:
        return True


class CallableOpInterface(OpTrait, abc.ABC):
    """
    Interface for function-like Operations that can be called in a generic way.

    Please see MLIR documentation for CallOpInterface and CallableOpInterface for more
    information.

    See external [documentation](https://mlir.llvm.org/docs/Interfaces/#callinterfaces).
    """

    @classmethod
    @abc.abstractmethod
    def get_callable_region(cls, op: Operation) -> Region:
        """
        Returns the body of the operation
        """
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def get_argument_types(cls, op: Operation) -> tuple[Attribute, ...]:
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def get_result_types(cls, op: Operation) -> tuple[Attribute, ...]:
        raise NotImplementedError()


@dataclass(frozen=True)
class HasCanonicalizationPatternsTrait(OpTrait):
    """
    Provides the rewrite passes to canonicalize an operation.

    Each rewrite pattern must have the trait's op as root.
    """

    def verify(self, op: Operation) -> None:
        return

    @classmethod
    @abc.abstractmethod
    def get_canonicalization_patterns(cls) -> tuple[RewritePattern, ...]:
        raise NotImplementedError()


@dataclass(frozen=True)
class HasShapeInferencePatternsTrait(OpTrait):
    """
    Provides the rewrite passes to shape infer an operation.

    Each rewrite pattern must have the trait's op as root.
    """

    def verify(self, op: Operation) -> None:
        return

    @classmethod
    @abc.abstractmethod
    def get_shape_inference_patterns(cls) -> tuple[RewritePattern, ...]:
        raise NotImplementedError()


class MemoryEffectKind(Enum):
    """
    The kind of side effect an operation can have.

    MLIR has a more detailed version of this, able to tie effects to specfic resources or
    values. Here, everything has its effect on the universe.
    """

    READ = auto()
    """
    Indicates that the operation reads from some resource. A 'read' effect implies only
    dereferencing of the resource, and not any visible mutation.
    """

    WRITE = auto()
    """
    Indicates that the operation writes to some resource. A 'write' effect implies only
    mutating a resource, and not any visible dereference or read.
    """

    ALLOC = auto()
    """
    Indicates that the operation allocates from some resource. An 'allocate' effect
    implies only allocation of the resource, and not any visible mutation or dereference.
    """

    FREE = auto()
    """
    Indicates that the operation frees some resource that has been allocated. A 'free'
    effect implies only de-allocation of the resource, and not any visible allocation,
    mutation or dereference.
    """


@dataclass(frozen=True)
class EffectInstance:
    """
    An instance of a side effect.
    """

    kind: MemoryEffectKind
    """
    The kind of side effect.
    """

    value: SSAValue | SymbolRefAttr | None = field(default=None)
    """
    The value or symbol that is affected by the side effect, if known.
    """


class MemoryEffect(OpTrait):
    """
    A trait that enables operations to expose their side-effects or absence thereof.
    """

    @classmethod
    @abc.abstractmethod
    def get_effects(cls, op: Operation) -> set[EffectInstance] | None:
        """
        Returns the concrete side effects of the operation.

        Return None if the operation cannot conclude - interpreted as if the operation
        had no MemoryEffect interface in the first place.
        """
        raise NotImplementedError()

    @final
    @classmethod
    def has_effects(cls, op: Operation) -> bool:
        """
        Returns if the operation has any side effects.
        """
        effects = cls.get_effects(op)
        return (effects is not None) and len(effects) > 0


def has_exact_effect(op: Operation, effect: MemoryEffectKind) -> bool:
    """
    Returns if the operation has the given side effects and no others.

    proxy for only_has_effect
    """
    return only_has_effect(op, effect)


def only_has_effect(op: Operation, effect: MemoryEffectKind) -> bool:
    """
    Returns if the operation has the given side effects and no others.
    """
    effects = get_effects(op)
    return effects is not None and all(e.kind == effect for e in effects)


def is_side_effect_free(op: Operation) -> bool:
    """
    Boilerplate helper to check if a generic operation is side effect free for sure.
    """
    effects = get_effects(op)
    return effects is not None and len(effects) == 0


def get_effects(op: Operation) -> set[EffectInstance] | None:
    """
    Helper to get known side effects of an operation.
    None means that the operation has unknown effects, for safety.
    """

    effect_interfaces = op.get_traits_of_type(MemoryEffect)
    if not effect_interfaces:
        return None

    effects = set[EffectInstance]()
    for it in op.get_traits_of_type(MemoryEffect):
        it_effects = it.get_effects(op)
        if it_effects is None:
            return None
        effects.update(it_effects)

    return effects


class NoMemoryEffect(MemoryEffect):
    """
    A trait that signals that an operation never has side effects.
    """

    @classmethod
    def get_effects(cls, op: Operation) -> set[EffectInstance]:
        return set()


class MemoryReadEffect(MemoryEffect):
    """
    A trait that signals that an operation always has read side effects.
    """

    @classmethod
    def get_effects(cls, op: Operation) -> set[EffectInstance]:
        return {EffectInstance(MemoryEffectKind.READ)}


class MemoryWriteEffect(MemoryEffect):
    """
    A trait that signals that an operation always has write side effects.
    """

    @classmethod
    def get_effects(cls, op: Operation) -> set[EffectInstance]:
        return {EffectInstance(MemoryEffectKind.WRITE)}


class MemoryAllocEffect(MemoryEffect):
    """
    A trait that signals that an operation always has alloc side effects.
    """

    @classmethod
    def get_effects(cls, op: Operation) -> set[EffectInstance]:
        return {EffectInstance(MemoryEffectKind.ALLOC)}


class MemoryFreeEffect(MemoryEffect):
    """
    A trait that signals that an operation always has deallocation side effects.
    """

    @classmethod
    def get_effects(cls, op: Operation) -> set[EffectInstance]:
        return {EffectInstance(MemoryEffectKind.FREE)}


class RecursiveMemoryEffect(MemoryEffect):
    """
    A trait that signals that an operation has the side effects of its contained
    operations.
    """

    @classmethod
    def get_effects(cls, op: Operation):
        effects = set[EffectInstance]()
        for r in op.regions:
            for b in r.blocks:
                for child_op in b.ops:
                    child_effects = get_effects(child_op)
                    if child_effects is None:
                        return None
                    effects.update(child_effects)
        return effects


class ConditionallySpeculatable(OpTrait):
    @classmethod
    @abc.abstractmethod
    def is_speculatable(cls, op: Operation) -> bool:
        raise NotImplementedError()


class AlwaysSpeculatable(ConditionallySpeculatable):
    @classmethod
    def is_speculatable(cls, op: Operation):
        return True


class RecursivelySpeculatable(ConditionallySpeculatable):
    @classmethod
    def is_speculatable(cls, op: Operation):
        return all(
            is_speculatable(o) for r in op.regions for b in r.blocks for o in b.ops
        )


def is_speculatable(op: Operation):
    trait = op.get_trait(ConditionallySpeculatable)
    return (trait is not None) and trait.is_speculatable(op)


class Pure(NoMemoryEffect, AlwaysSpeculatable):
    """
    In MLIR, Pure is NoMemoryEffect + AlwaysSpeculatable, but the latter is nowhere to be
    found here.
    """


class Commutative(OpTrait):
    """
    A trait that signals that an operation is commutative.
    """


class HasInsnRepresentation(OpTrait, abc.ABC):
    """
    A trait providing information on how to encode an operation using a .insn assember directive.

    The returned string contains python string.format placeholders where formatted operands are inserted during
    printing.

    See external [documentation](https://sourceware.org/binutils/docs/as/RISC_002dV_002dDirectives.html for more information.).
    """

    @abc.abstractmethod
    def get_insn(self, op: Operation) -> str:
        """
        Return the insn representation of the operation for printing.
        """
        raise NotImplementedError()


@dataclass(frozen=True)
class SameOperandsAndResultType(OpTrait):
    """Constrain the operation to have the same operands and result type."""

    def verify(self, op: Operation) -> None:
        from xdsl.utils.type import (
            get_element_type_or_self,
            get_encoding,
            have_compatible_shape,
        )

        if len(op.results) < 1 or len(op.operands) < 1:
            raise VerifyException(
                f"'{op.name}' requires at least one result or operand"
            )

        result_type0 = get_element_type_or_self(op.result_types[0])

        encoding = get_encoding(op.result_types[0])

        for result_type in op.result_types[1:]:
            result_type_elem = get_element_type_or_self(result_type)
            if result_type0 != result_type_elem or not have_compatible_shape(
                op.result_types[0], result_type
            ):
                raise VerifyException(
                    f"'{op.name} requires the same type for all operands and results"
                )

            element_encoding = get_encoding(result_type)

            if encoding != element_encoding:
                raise VerifyException(
                    f"'{op.name} requires the same encoding for all operands and results"
                )

        for operand_type in op.operand_types:
            operand_type_elem = get_element_type_or_self(operand_type)
            if result_type0 != operand_type_elem or not have_compatible_shape(
                op.result_types[0], operand_type
            ):
                raise VerifyException(
                    f"'{op.name} requires the same type for all operands and results"
                )

            element_encoding = get_encoding(operand_type)

            if encoding != element_encoding:
                raise VerifyException(
                    f"'{op.name} requires the same encoding for all operands and results"
                )



import os
from dataclasses import dataclass

from xdsl.context import Context
from xdsl.dialects import builtin, pdl_interp
from xdsl.dialects.builtin import ModuleOp
from xdsl.interpreter import Interpreter
from xdsl.interpreters.pdl_interp import PDLInterpFunctions
from xdsl.ir import Operation
from xdsl.parser import Parser
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import PatternRewriter, PatternRewriteWalker, RewritePattern


@dataclass
class PDLInterpRewritePattern(RewritePattern):
    """
    A rewrite pattern that uses the pdl_interp dialect for matching and rewriting operations.
    """

    interpreter: Interpreter
    functions: PDLInterpFunctions
    matcher: pdl_interp.FuncOp

    def __init__(
        self,
        matcher: pdl_interp.FuncOp,
        interpreter: Interpreter,
        functions: PDLInterpFunctions,
    ):
        self.functions = functions
        module = matcher.parent_op()
        assert isinstance(module, ModuleOp)
        self.interpreter = interpreter
        if matcher.sym_name.data != "matcher":
            raise ValueError("Matcher function name must be 'matcher'")
        self.matcher = matcher

    def match_and_rewrite(self, xdsl_op: Operation, rewriter: PatternRewriter) -> None:
        # Setup the rewriter
        self.functions.rewriter = rewriter

        # Call the matcher function on the operation
        self.interpreter.call_op(self.matcher, (xdsl_op,))


@dataclass(frozen=True)
class ApplyPDLInterpPass(ModulePass):
    name = "apply-pdl-interp"

    pdl_interp_file: str | None = None

    def apply(self, ctx: Context, op: builtin.ModuleOp) -> None:
        if self.pdl_interp_file is not None:
            assert os.path.exists(self.pdl_interp_file)
            with open(self.pdl_interp_file) as f:
                pdl_interp_module_str = f.read()
                parser = Parser(ctx, pdl_interp_module_str)
                pdl_interp_module = parser.parse_module()
        else:
            pdl_interp_module = op
        matcher = None
        for cur in pdl_interp_module.walk():
            if isinstance(cur, pdl_interp.FuncOp) and cur.sym_name.data == "matcher":
                matcher = cur
                break
        assert matcher is not None, "matcher function not found"
        interpreter = Interpreter(pdl_interp_module)
        implementations = PDLInterpFunctions(ctx)
        interpreter.register_implementations(implementations)
        rewrite_pattern = PDLInterpRewritePattern(matcher, interpreter, implementations)
        PatternRewriteWalker(rewrite_pattern).rewrite_module(op)



from typing import Any

from xdsl.builder import Builder, ImplicitBuilder
from xdsl.dialects import affine, arith, func, memref
from xdsl.dialects.builtin import AffineMapAttr, IndexType, ModuleOp
from xdsl.interpreter import Interpreter
from xdsl.interpreters.affine import AffineFunctions
from xdsl.interpreters.arith import ArithFunctions
from xdsl.interpreters.func import FuncFunctions
from xdsl.interpreters.memref import MemRefFunctions
from xdsl.interpreters.shaped_array import ShapedArray
from xdsl.interpreters.utils.ptr import TypedPtr
from xdsl.ir.affine import AffineMap
from xdsl.utils.test_value import create_ssa_value

index = IndexType()


@ModuleOp
@Builder.implicit_region
def module_op():
    with ImplicitBuilder(
        func.FuncOp(
            "my_func",
            (
                (),
                (
                    memref.MemRefType(index, (2, 3)),
                    memref.MemRefType(index, (3, 2)),
                ),
            ),
        ).body
    ):
        alloc_op_0 = memref.AllocOp.get(index, None, (2, 3))
        alloc_op_1 = memref.AllocOp.get(index, None, (3, 2))

        @Builder.implicit_region((index,))
        def init_rows_region(args: tuple[Any, ...]):
            (row,) = args

            @Builder.implicit_region((index,))
            def init_cols_region(args: tuple[Any, ...]):
                (col,) = args
                sum_op = arith.AddiOp(row, col)
                affine.StoreOp(sum_op.result, alloc_op_0.memref, (row, col))
                affine.YieldOp.get()

            affine.ForOp.from_region((), (), (), (), 0, 3, init_cols_region)

            affine.YieldOp.get()

        affine.ForOp.from_region((), (), (), (), 0, 2, init_rows_region)

        @Builder.implicit_region((index,))
        def transpose_rows_region(args: tuple[Any, ...]):
            (row,) = args

            @Builder.implicit_region((index,))
            def transpose_cols_region(args: tuple[Any, ...]):
                (col,) = args
                res = affine.LoadOp(alloc_op_0.memref, (row, col)).result
                affine.StoreOp(res, alloc_op_1.memref, (col, row))
                affine.YieldOp.get()

            affine.ForOp.from_region((), (), (), (), 0, 3, transpose_cols_region)

            affine.YieldOp.get()

        affine.ForOp.from_region((), (), (), (), 0, 2, transpose_rows_region)
        func.ReturnOp(alloc_op_0.memref, alloc_op_1.memref)


def test_functions():
    module_op.verify()

    interpreter = Interpreter(module_op)
    interpreter.register_implementations(ArithFunctions())
    interpreter.register_implementations(MemRefFunctions())
    interpreter.register_implementations(AffineFunctions())
    interpreter.register_implementations(FuncFunctions())

    res = interpreter.call_op("my_func", ())

    assert res == (
        ShapedArray(
            TypedPtr.new_index([0, 1, 2, 1, 2, 3], interpreter.index_bitwidth), [2, 3]
        ),
        ShapedArray(
            TypedPtr.new_index([0, 1, 1, 2, 2, 3], interpreter.index_bitwidth), [3, 2]
        ),
    )


def test_apply():
    interpreter = Interpreter(ModuleOp([]))
    interpreter.register_implementations(AffineFunctions())

    assert interpreter.run_op(
        affine.ApplyOp(
            (create_ssa_value(index), create_ssa_value(index)),
            AffineMapAttr(AffineMap.from_callable(lambda d0, d1: (d0 + d1,))),
        ),
        (1, 2),
    ) == (3,)



from pathlib import Path
from typing import cast

from xdsl.parser import GenericParser, ParserState
from xdsl.utils.lexer import Input

from .lexer import ToyLexer, ToyToken, ToyTokenKind
from .toy_ast import (
    BinaryExprAST,
    CallExprAST,
    ExprAST,
    FunctionAST,
    LiteralExprAST,
    ModuleAST,
    NumberExprAST,
    PrintExprAST,
    PrototypeAST,
    ReturnExprAST,
    VarDeclExprAST,
    VariableExprAST,
    VarType,
)


class ToyParser(GenericParser[ToyTokenKind]):
    def __init__(self, file: Path, program: str):
        super().__init__(ParserState(ToyLexer(Input(program, str(file)))))

    def get_token_precedence(self) -> int:
        """Returns precedence if the current token is a binary operation, -1 otherwise"""
        PRECEDENCE = {
            "-": 20,
            "+": 20,
            "*": 40,
        }
        op = self._current_token.text

        return PRECEDENCE.get(op, -1)

    def _peek(self, expected_kind: ToyTokenKind) -> ToyToken | None:
        """
        Returns token matching pattern or None.
        """
        token = self._current_token
        if token.kind == expected_kind:
            return token

    def _pop(self, expected_kind: ToyTokenKind) -> ToyToken:
        """
        Verifies that the current token is of expected type,
        raises ParseError otherwise.
        """
        return self._parse_token(expected_kind, f"Expected {expected_kind}")

    def parse_module(self):
        """
        Parse a full Module. A module is a list of function definitions.
        """
        functions: list[FunctionAST] = []

        while self._parse_optional_token(ToyTokenKind.EOF) is None:
            functions.append(self.parse_definition())

        return ModuleAST(tuple(functions))

    def parse_return(self):
        """
        Parse a return statement.
        return :== return ; | return expr ;
        """
        return_token = self._pop(ToyTokenKind.RETURN)
        expr = None

        # Return takes an optional argument
        if self._parse_optional_token(ToyTokenKind.SEMICOLON) is None:
            expr = self.parse_expression()

        return ReturnExprAST(return_token.span.get_location(), expr)

    def parse_number_expr(self):
        """
        Parse a literal number.
        numberexpr ::= number
        """
        number_token = self._pop(ToyTokenKind.NUMBER)
        return NumberExprAST(
            number_token.span.get_location(), float(number_token.span.text)
        )

    def parse_tensor_literal_expr(self) -> LiteralExprAST | NumberExprAST:
        """
        Parse a literal array expression.
        tensorLiteral ::= [ literalList ] | number
        literalList ::= tensorLiteral | tensorLiteral, literalList
        """
        if self._peek(ToyTokenKind.NUMBER):
            return self.parse_number_expr()

        open_bracket = self._current_token

        # Hold the list of values at this nesting level.
        values = self.parse_comma_separated_list(
            self.Delimiter.SQUARE, self.parse_tensor_literal_expr
        )

        # Hold the dimensions for all the nesting inside this level.
        # Fill in the dimensions now. First the current nesting level:
        dims = [len(values)]

        # If there is any nested array, process all of them and ensure
        # that dimensions are uniform.
        if any(type(val) is LiteralExprAST for val in values):
            all_tensors = all(type(val) is LiteralExprAST for val in values)
            if not all_tensors:
                self.raise_error(
                    "Expected uniform well-nested dimensions inside literal expression"
                )

            tensor_values = cast(list[LiteralExprAST], values)
            first = tensor_values[0].dims
            all_equal = all(val.dims == first for val in tensor_values)
            if not all_equal:
                self.raise_error(
                    "Expected uniform well-nested dimensions inside literal expression"
                )

            dims += first

        return LiteralExprAST(open_bracket.span.get_location(), values, dims)

    def parse_paren_expr(self) -> ExprAST:
        "parenexpr ::= '(' expression ')'"
        with self.in_parens():
            return self.parse_expression()

    def parse_identifier_expr(self):
        """
        identifierexpr
        ::= identifier
        ::= identifier '(' expression ')'
        """
        name = self._pop(ToyTokenKind.IDENTIFIER)
        args = self.parse_optional_comma_separated_list(
            self.Delimiter.PAREN, self.parse_expression
        )
        if args is None:
            # Simple variable ref.
            return VariableExprAST(name.span.get_location(), name.text)

        # This is a function call.
        if name.text == "print":
            # It can be a builtin call to print
            if len(args) != 1:
                self.raise_error("Expected <single arg> as argument to print()")

            return PrintExprAST(name.span.get_location(), args[0])

        return CallExprAST(name.span.get_location(), name.text, args)

    def parse_primary(self) -> ExprAST | None:
        """
        primary
        ::= identifierexpr
        ::= numberexpr
        ::= parenexpr
        ::= tensorliteral
        """
        current = self._current_token
        if current.kind == ToyTokenKind.IDENTIFIER:
            return self.parse_identifier_expr()
        elif current.kind == ToyTokenKind.NUMBER:
            return self.parse_number_expr()
        elif current.kind == ToyTokenKind.PARENTHESE_OPEN:
            return self.parse_paren_expr()
        elif current.kind == ToyTokenKind.SBRACKET_OPEN:
            return self.parse_tensor_literal_expr()
        elif current.kind == ToyTokenKind.SEMICOLON:
            return None
        elif current.kind == ToyTokenKind.BRACKET_CLOSE:
            return None
        else:
            self.raise_error("Expected expression or one of `;`, `}`")

    def parse_primary_not_none(self) -> ExprAST:
        """
        primary
        ::= identifierexpr
        ::= numberexpr
        ::= parenexpr
        ::= tensorliteral
        """
        current = self._current_token
        if current.kind == ToyTokenKind.IDENTIFIER:
            return self.parse_identifier_expr()
        elif current.kind == ToyTokenKind.NUMBER:
            return self.parse_number_expr()
        elif current.kind == ToyTokenKind.PARENTHESE_OPEN:
            return self.parse_paren_expr()
        elif current.kind == ToyTokenKind.SBRACKET_OPEN:
            return self.parse_tensor_literal_expr()
        else:
            self.raise_error("Expected expression")

    def parse_bin_op_rhs(self, expr_precedence: int, lhs: ExprAST) -> ExprAST:
        """
        Recursively parse the right hand side of a binary expression, the ExprPrec
        argument indicates the precedence of the current binary operator.

        binoprhs ::= ('+' primary)*
        """
        # If this is a binop, find its precedence.
        while True:
            tok_precedence = self.get_token_precedence()

            # If this is a binop that binds at least as tightly as the current binop,
            # consume it, otherwise we are done.
            if tok_precedence < expr_precedence:
                return lhs

            # Okay, we know this is a binop.
            binOp = self._pop(ToyTokenKind.OPERATOR).text

            # Parse the primary expression after the binary operator.
            rhs = self.parse_primary()

            if rhs is None:
                self.raise_error("Expected expression to complete binary operator")

            # If BinOp binds less tightly with rhs than the operator after rhs, let
            # the pending operator take rhs as its lhs.
            next_precedence = self.get_token_precedence()
            if tok_precedence < next_precedence:
                rhs = self.parse_bin_op_rhs(tok_precedence + 1, rhs)

            # Merge lhs/rhs
            lhs = BinaryExprAST(rhs.loc, binOp, lhs, rhs)

    def parse_expression(self) -> ExprAST:
        """expression::= primary binop rhs"""
        lhs = self.parse_primary_not_none()
        return self.parse_bin_op_rhs(0, lhs)

    def parse_type(self):
        """
        type ::= < shape_list >
        shape_list ::= num | num , shape_list
        """
        return VarType(
            self.parse_comma_separated_list(
                self.Delimiter.ANGLE,
                lambda: int(self._pop(ToyTokenKind.NUMBER).text),
            )
        )

    def parse_declaration(self):
        """
        Parse a variable declaration, it starts with a `var` keyword followed by
        and identifier and an optional type (shape specification) before the
        initializer.
        decl ::= var identifier [ type ] = expr
        """
        var = self._pop(ToyTokenKind.VAR)
        name = self._pop(ToyTokenKind.IDENTIFIER).text
        var_type = self.parse_type() if self._peek(ToyTokenKind.LT) else VarType([])
        self._pop(ToyTokenKind.EQ)
        expr = self.parse_expression()
        return VarDeclExprAST(var.span.get_location(), name, var_type, expr)

    def parse_block(self) -> tuple[ExprAST, ...]:
        """
        Parse a block: a list of expression separated by semicolons and wrapped in
        curly braces.

        block ::= { expression_list }
        expression_list ::= block_expr ; expression_list
        block_expr ::= decl | "return" | expr
        """
        self._pop(ToyTokenKind.BRACKET_OPEN)
        exprs: list[ExprAST] = []

        # Ignore empty expressions: swallow sequences of semicolons.
        while self._parse_optional_token(ToyTokenKind.SEMICOLON):
            continue

        while not self._parse_optional_token(ToyTokenKind.BRACKET_CLOSE):
            if self._peek(ToyTokenKind.VAR):
                # Variable declaration
                exprs.append(self.parse_declaration())
            elif self._peek(ToyTokenKind.RETURN):
                # Return statement
                exprs.append(self.parse_return())
            else:
                # General expression
                exprs.append(self.parse_expression())

            # Ensure that elements are separated by a semicolon.
            self._pop(ToyTokenKind.SEMICOLON)

            # Ignore empty expressions: swallow sequences of semicolons.
            while self._parse_optional_token(ToyTokenKind.SEMICOLON):
                continue

        return tuple(exprs)

    def _parse_arg(self) -> VariableExprAST:
        arg = self._pop(ToyTokenKind.IDENTIFIER)
        return VariableExprAST(arg.span.get_location(), arg.text)

    def parse_prototype(self):
        """
        prototype ::= def id '(' decl_list ')'
        decl_list ::= identifier | identifier, decl_list
        """
        def_token = self._pop(ToyTokenKind.DEF)
        name = self._pop(ToyTokenKind.IDENTIFIER).text
        args = self.parse_comma_separated_list(self.Delimiter.PAREN, self._parse_arg)
        return PrototypeAST(def_token.span.get_location(), name, args)

    def parse_definition(self):
        """
        Parse a function definition, we expect a prototype initiated with the
        `def` keyword, followed by a block containing a list of expressions.

        definition ::= prototype block
        """
        proto = self.parse_prototype()
        block = self.parse_block()
        return FunctionAST(proto.loc, proto, block)
